import base64
import json
import unittest

from django.core.files.uploadedfile import SimpleUploadedFile
from django.test.client import RequestFactory
from djblets.util.filesystem import is_exe_in_path
from kgb import SpyAgency

from reviewboard.diffviewer.diffutils import (get_original_file,
                                              get_patched_file,
                                              patch)
from reviewboard.diffviewer.errors import (DiffParserError, DiffTooBigError,
                                           EmptyDiffError)
from reviewboard.diffviewer.forms import (UploadCommitForm, UploadDiffForm,
                                          ValidateCommitForm)
from reviewboard.diffviewer.models import DiffSet, DiffSetHistory
from reviewboard.scmtools.errors import FileNotFoundError
from reviewboard.scmtools.models import Tool
from reviewboard.testing import TestCase


class UploadCommitFormTests(SpyAgency, TestCase):
    """Unit tests for UploadCommitForm."""

    fixtures = ['test_scmtools']

    _default_form_data = {
        'base_commit_id': '1234',
        'basedir': '/',
        'commit_id': 'r1',
        'parent_id': 'r0',
        'commit_message': 'Message',
        'author_name': 'Author',
        'author_email': 'author@example.org',
        'author_date': '1970-01-01 00:00:00+0000',
        'committer_name': 'Committer',
        'committer_email': 'committer@example.org',
        'committer_date': '1970-01-01 00:00:00+0000',
    }

    def setUp(self):
        super(UploadCommitFormTests, self).setUp()

        self.repository = self.create_repository(tool_name='Git')
        self.spy_on(self.repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)
        self.diffset = DiffSet.objects.create_empty(repository=self.repository)

    def test_create(self):
        """Testing UploadCommitForm.create"""
        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                  content_type='text/x-patch')

        form = UploadCommitForm(
            diffset=self.diffset,
            data=self._default_form_data.copy(),
            files={
                'diff': diff,
            })

        self.assertTrue(form.is_valid())
        commit = form.create()

        self.assertEqual(self.diffset.files.count(), 1)
        self.assertEqual(self.diffset.commits.count(), 1)
        self.assertEqual(commit.files.count(), 1)
        self.assertEqual(set(self.diffset.files.all()),
                         set(commit.files.all()))

    def test_clean_parent_diff_path(self):
        """Testing UploadCommitForm.clean() for a subsequent commit with a
        parent diff
        """
        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                  content_type='text/x-patch')
        parent_diff = SimpleUploadedFile('parent_diff',
                                         self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                         content_type='text/x-patch')

        form = UploadCommitForm(
            diffset=self.diffset,
            data=self._default_form_data.copy(),
            files={
                'diff': diff,
                'parent_diff': parent_diff,
            })

        self.assertTrue(form.is_valid())
        form.create()

        form = UploadCommitForm(
            diffset=self.diffset,
            data=dict(
                self._default_form_data,
                **{
                    'parent_id': 'r1',
                    'commit_id': 'r2',
                }
            ),
            files={
                'diff': diff,
                'parent_diff': parent_diff,
            })

        self.assertTrue(form.is_valid())
        self.assertNotIn('parent_diff', form.errors)

    def test_clean_published_diff(self):
        """Testing UploadCommitForm.clean() for a DiffSet that has already been
        published
        """
        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                  content_type='text/x-patch')

        form = UploadCommitForm(
            diffset=self.diffset,
            data=self._default_form_data,
            files={
                'diff': diff,
            })

        self.assertTrue(form.is_valid())
        form.create()

        self.diffset.history = DiffSetHistory.objects.create()
        self.diffset.save(update_fields=('history_id',))

        form = UploadCommitForm(
            diffset=self.diffset,
            data=dict(
                self._default_form_data,
                parent_id='r1',
                commit_id='r0',
            ),
            files={
                'diff_path': SimpleUploadedFile(
                    'diff',
                    self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                    content_type='text/x-patch'),
            })

        self.assertFalse(form.is_valid())
        self.assertNotEqual(form.non_field_errors, [])

    def test_clean_author_date(self):
        """Testing UploadCommitForm.clean_author_date"""
        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                  content_type='text/x-patch')

        form = UploadCommitForm(
            diffset=self.diffset,
            data=dict(self._default_form_data, **{
                'author_date': 'Jan 1 1970',
            }),
            files={
                'diff': diff,
            })

        self.assertFalse(form.is_valid())
        self.assertIn('author_date', form.errors)

    def test_clean_committer_date(self):
        """Testing UploadCommitForm.clean_committer_date"""
        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                  content_type='text/x-patch')

        form = UploadCommitForm(
            diffset=self.diffset,
            data=dict(self._default_form_data, **{
                'committer_date': 'Jun 1 1970',
            }),
            files={
                'diff': diff,
            })

        self.assertFalse(form.is_valid())
        self.assertIn('committer_date', form.errors)

    def test_clean_no_committer(self):
        """Testing UploadCommitForm.clean when no committer_ fields are present
        """
        field_names = {
            'committer_date',
            'committer_email',
            'committer_name',
        }

        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                  content_type='text/x-patch')

        form_data = self._default_form_data.copy()

        for field in field_names:
            del form_data[field]

        form = UploadCommitForm(
            diffset=self.diffset,
            data=form_data,
            files={
                'diff': diff,
            })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {
            field: ['This field is required.']
            for field in field_names
        })

    @unittest.skipIf(not is_exe_in_path('hg'),
                     'Hg is not installed')
    def test_clean_committer_unsupported(self):
        """Testing UploadCommitForm.clean when committer_ fields are present
        for a SCMTool that doesn't support them
        """
        self.repository.tool = Tool.objects.get(name='Mercurial')
        self.repository.scmtool_id = 'mercurial'
        self.repository.save(update_fields=('tool', 'scmtool_id'))

        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                  content_type='text/x-patch')

        form = UploadCommitForm(
            diffset=self.diffset,
            data=self._default_form_data.copy(),
            files={
                'diff': diff,
            })

        self.assertTrue(form.is_valid())

        self.assertNotIn('committer_date', form.cleaned_data)
        self.assertNotIn('committer_email', form.cleaned_data)
        self.assertNotIn('committer_name', form.cleaned_data)


class UploadDiffFormTests(SpyAgency, TestCase):
    """Unit tests for UploadDiffForm."""

    fixtures = ['test_scmtools']

    def test_create(self):
        """Testing UploadDiffForm.create"""
        diff_file = SimpleUploadedFile('diff',
                                       self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                       content_type='text/x-patch')

        repository = self.create_repository(tool_name='Test')

        self.spy_on(repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)

        form = UploadDiffForm(
            repository=repository,
            data={
                'basedir': '/',
                'base_commit_id': '1234',
            },
            files={
                'path': diff_file,
            })
        self.assertTrue(form.is_valid())

        diffset = form.create()
        self.assertEqual(diffset.files.count(), 1)
        self.assertEqual(diffset.basedir, '/')
        self.assertEqual(diffset.base_commit_id, '1234')

    def test_create_filters_parent_diffs(self):
        """Testing UploadDiffForm.create filters parent diff files"""
        saw_file_exists = {}

        def get_file_exists(repository, path, revision, *args, **kwargs):
            saw_file_exists[(path, revision)] = True
            return True

        parent_diff_1 = (
            b'diff --git a/README b/README\n'
            b'index d6613f4..5b50865 100644\n'
            b'--- README\n'
            b'+++ README\n'
            b'@@ -2 +2 @@\n'
            b'-blah..\n'
            b'+blah blah\n'
        )
        parent_diff_2 = (
            b'diff --git a/UNUSED b/UNUSED\n'
            b'index 1234567..5b50866 100644\n'
            b'--- UNUSED\n'
            b'+++ UNUSED\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-foo\n'
            b'+bar\n'
        )
        parent_diff = parent_diff_1 + parent_diff_2

        diff_file = SimpleUploadedFile('diff',
                                       self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                       content_type='text/x-patch')
        parent_diff_file = SimpleUploadedFile('parent_diff', parent_diff,
                                              content_type='text/x-patch')

        repository = self.create_repository(tool_name='Test')
        self.spy_on(repository.get_file_exists, call_fake=get_file_exists)

        form = UploadDiffForm(
            repository=repository,
            data={
                'basedir': '/',
            },
            files={
                'path': diff_file,
                'parent_diff_path': parent_diff_file,
            })
        self.assertTrue(form.is_valid())

        diffset = form.create()
        self.assertEqual(diffset.files.count(), 1)

        filediff = diffset.files.get()
        self.assertEqual(filediff.diff, self.DEFAULT_GIT_FILEDIFF_DATA_DIFF)
        self.assertEqual(filediff.parent_diff, parent_diff_1)

        self.assertIn(('/README', 'd6613f4'), saw_file_exists)
        self.assertNotIn(('/UNUSED', '1234567'), saw_file_exists)
        self.assertEqual(len(saw_file_exists), 1)

    @unittest.skipIf(not is_exe_in_path('hg'),
                     'Hg is not installed')
    def test_create_with_parser_get_orig_commit_id(self):
        """Testing UploadDiffForm.create uses correct base revision set in
        ParsedDiffChange.parent_commit_id
        """
        diff = (
            b'# Node ID a6fc203fee9091ff9739c9c00cd4a6694e023f48\n'
            b'# Parent  7c4735ef51a7c665b5654f1a111ae430ce84ebbd\n'
            b'diff --git a/doc/readme b/doc/readme\n'
            b'--- a/doc/readme\n'
            b'+++ b/doc/readme\n'
            b'@@ -1,3 +1,3 @@\n'
            b' Hello\n'
            b'-\n'
            b'+...\n'
            b' goodbye\n'
        )

        parent_diff = (
            b'# Node ID 7c4735ef51a7c665b5654f1a111ae430ce84ebbd\n'
            b'# Parent  661e5dd3c4938ecbe8f77e2fdfa905d70485f94c\n'
            b'diff --git a/doc/newfile b/doc/newfile\n'
            b'new file mode 100644\n'
            b'--- /dev/null\n'
            b'+++ b/doc/newfile\n'
            b'@@ -0,0 +1,1 @@\n'
            b'+Lorem ipsum\n'
        )

        diff_file = SimpleUploadedFile('diff', diff,
                                       content_type='text/x-patch')
        parent_diff_file = SimpleUploadedFile('parent_diff', parent_diff,
                                              content_type='text/x-patch')

        repository = self.create_repository(
            name='Test HG',
            path='scmtools/testdata/hg_repo',
            tool_name='Mercurial')

        form = UploadDiffForm(
            repository=repository,
            files={
                'path': diff_file,
                'parent_diff_path': parent_diff_file,
            })
        self.assertTrue(form.is_valid())

        diffset = form.create()
        self.assertEqual(diffset.files.count(), 1)

        filediff = diffset.files.get()

        self.assertEqual(filediff.source_revision,
                         '7c4735ef51a7c665b5654f1a111ae430ce84ebbd')
        self.assertEqual(filediff.extra_data.get('parent_source_revision'),
                         '661e5dd3c4938ecbe8f77e2fdfa905d70485f94c')

    def test_create_with_parent_filediff_with_move_and_no_change(self):
        """Testing UploadDiffForm.create with a parent diff consisting only
        of a move/rename without content change
        """
        revisions = [
            b'93e6b3e8944c48737cb11a1e52b046fa30aea7a9',
            b'4839fc480f47ca59cf05a9c39410ea744d1e17a2',
        ]

        parent_diff = SimpleUploadedFile(
            'parent_diff',
            (b'diff --git a/foo b/bar\n'
             b'similarity index 100%%\n'
             b'rename from foo\n'
             b'rename to bar\n'),
            content_type='text/x-patch')

        diff = SimpleUploadedFile(
            'diff',
            (b'diff --git a/bar b/bar\n'
             b'index %s..%s 100644\n'
             b'--- a/bar\n'
             b'+++ b/bar\n'
             b'@@ -1,1 +1,2 @@\n'
             b' Foo\n'
             b'+Bar\n') % (revisions[0], revisions[1]),
            content_type='text/x-patch')

        repository = self.create_repository(tool_name='Test')
        self.spy_on(repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)

        # We will only be making one call to get_file and we can fake it out.
        self.spy_on(repository.get_file,
                    call_fake=lambda *args, **kwargs: b'Foo\n')
        self.spy_on(patch)

        form = UploadDiffForm(
            repository=repository,
            data={
                'basedir': '/',
            },
            files={
                'path': diff,
                'parent_diff_path': parent_diff,
            })
        self.assertTrue(form.is_valid())

        diffset = form.create()
        self.assertEqual(diffset.files.count(), 1)

        f = diffset.files.get()
        self.assertEqual(f.source_revision, revisions[0].decode('utf-8'))
        self.assertEqual(f.dest_detail, revisions[1].decode('utf-8'))

        # We shouldn't call out to patch because the parent diff is just a
        # rename.
        original_file = get_original_file(filediff=f,
                                          request=None)
        self.assertEqual(original_file, b'Foo\n')
        self.assertFalse(patch.spy.called)

        patched_file = get_patched_file(source_data=original_file,
                                        filediff=f)
        self.assertEqual(patched_file, b'Foo\nBar\n')
        self.assertTrue(patch.spy.called)

    def test_create_with_parent_filediff_with_move_and_change(self):
        """Testing UploadDiffForm.create with a parent diff consisting of a
        move/rename with content change
        """
        revisions = [
            b'5d36b88bb697a2d778f024048bafabd443d74503',
            b'9b32edcd37a88c6ada91efc562afa637ccfdad36',
            b'8a567d328293f85d68332bc693b0a98869b23b47',
        ]

        parent_diff = SimpleUploadedFile(
            'parent_diff',
            (b'diff --git a/foo b/bar\n'
             b'similarity index 55%%\n'
             b'rename from foo\n'
             b'rename to bar\n'
             b'index %s..%s 100644\n'
             b'--- a/foo\n'
             b'+++ b/bar\n'
             b'@@ -1,1 +1,2 @@\n'
             b' Foo\n'
             b'+Bar\n') % (revisions[0], revisions[1]),
            content_type='text/x-patch')

        diff = SimpleUploadedFile(
            'diff',
            (b'diff --git a/bar b/bar\n'
             b'index %s..%s 100644\n'
             b'--- a/bar\n'
             b'+++ b/bar\n'
             b'@@ -1,2 +1,3 @@\n'
             b' Foo\n'
             b' Bar\n'
             b'+Baz\n') % (revisions[1], revisions[2]),
            content_type='text/x-patch')

        repository = self.create_repository(tool_name='Test')
        self.spy_on(repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)

        # We will only be making one call to get_file and we can fake it out.
        self.spy_on(repository.get_file,
                    call_fake=lambda *args, **kwargs: b'Foo\n')
        self.spy_on(patch)

        form = UploadDiffForm(
            repository=repository,
            data={
                'basedir': '/',
            },
            files={
                'path': diff,
                'parent_diff_path': parent_diff,
            })
        self.assertTrue(form.is_valid())

        diffset = form.create()
        self.assertEqual(diffset.files.count(), 1)

        filediff = diffset.files.get()
        self.assertEqual(filediff.source_file, 'bar')
        self.assertEqual(filediff.dest_file, 'bar')
        self.assertEqual(filediff.source_revision,
                         revisions[1].decode('utf-8'))
        self.assertEqual(filediff.dest_detail, revisions[2].decode('utf-8'))
        self.assertEqual(filediff.extra_data, {
            '__parent_diff_empty': False,
            'is_symlink': False,
            'new_unix_mode': '100644',
            'old_unix_mode': '100644',
            'parent_moved': True,
            'parent_source_filename': '/foo',
            'parent_source_revision': revisions[0].decode('utf-8'),
            'raw_delete_count': 0,
            'raw_insert_count': 1,
        })

        original_file = get_original_file(filediff=filediff,
                                          request=None)
        self.assertEqual(original_file, b'Foo\nBar\n')
        self.assertTrue(patch.spy.called)

        patched_file = get_patched_file(source_data=original_file,
                                        filediff=filediff)
        self.assertEqual(patched_file, b'Foo\nBar\nBaz\n')
        self.assertEqual(len(patch.spy.calls), 2)

    def test_create_missing_basedir(self):
        """Testing UploadDiffForm with a missing basedir field that is
        required
        """
        repository = self.create_repository(tool_name='Test')
        scmtool = repository.get_scmtool()

        self.spy_on(repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)

        revisions = [
            b'93e6b3e8944c48737cb11a1e52b046fa30aea7a9',
            b'4839fc480f47ca59cf05a9c39410ea744d1e17a2',
        ]

        diff = SimpleUploadedFile(
            'diff',
            (b'diff --git a/bar b/bar\n'
             b'index %s..%s 100644\n'
             b'--- a/bar\n'
             b'+++ b/bar\n'
             b'@@ -1,2 +1,3 @@\n'
             b' Foo\n'
             b'+Bar\n') % (revisions[0], revisions[1]),
            content_type='text/x-patch')

        orig_use_abs_paths = scmtool.diffs_use_absolute_paths

        try:
            scmtool.diffs_use_absolute_paths = True

            form = UploadDiffForm(
                repository=repository,
                files={
                    'path': diff,
                }
            )

            self.assertFalse(form.is_valid())
        finally:
            scmtool.diffs_use_absolute_paths = orig_use_abs_paths

        self.assertIn('basedir', form.errors)
        self.assertIn('This field is required.', form.errors['basedir'])

    def test_create_with_parent_filediff_with_new_file(self):
        """Testing UploadDiffForm.create with a parent diff consisting of a
        newly-introduced file
        """
        revisions = [
            b'0000000000000000000000000000000000000000',
            b'9b32edcd37a88c6ada91efc562afa637ccfdad36',
            b'8a567d328293f85d68332bc693b0a98869b23b47',
        ]

        parent_diff = SimpleUploadedFile(
            'parent_diff',
            (b'diff --git a/foo b/foo\n'
             b'new file mode 100644\n'
             b'index %s..%s\n'
             b'--- /dev/null\n'
             b'+++ b/foo\n'
             b'@@ -0,0 +1,2 @@\n'
             b'+Foo\n'
             b'+Bar\n') % (revisions[0], revisions[1]),
            content_type='text/x-patch')

        diff = SimpleUploadedFile(
            'diff',
            (b'diff --git a/foo b/foo\n'
             b'index %s..%s 100644\n'
             b'--- a/foo\n'
             b'+++ b/foo\n'
             b'@@ -1,3 +1,4 @@\n'
             b' Foo\n'
             b' Bar\n'
             b'+Baz\n') % (revisions[1], revisions[2]),
            content_type='text/x-patch')

        repository = self.create_repository(tool_name='Test')
        self.spy_on(repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)

        # We will only be making one call to get_file and we can fake it out.
        self.spy_on(repository.get_file,
                    call_fake=lambda *args, **kwargs: b'Foo\n')
        self.spy_on(patch)

        form = UploadDiffForm(
            repository=repository,
            data={
                'basedir': '/',
            },
            files={
                'parent_diff_path': parent_diff,
                'path': diff,
            })
        self.assertTrue(form.is_valid())

        diffset = form.create()
        self.assertEqual(diffset.files.count(), 1)

        filediff = diffset.files.get()
        self.assertEqual(filediff.source_file, 'foo')
        self.assertEqual(filediff.dest_file, 'foo')
        self.assertEqual(filediff.source_revision,
                         revisions[1].decode('utf-8'))
        self.assertEqual(filediff.dest_detail, revisions[2].decode('utf-8'))
        self.assertEqual(filediff.extra_data, {
            '__parent_diff_empty': False,
            'is_symlink': False,
            'new_unix_mode': '100644',
            'old_unix_mode': '100644',
            'parent_source_filename': '/foo',
            'parent_source_revision': 'PRE-CREATION',
            'raw_delete_count': 0,
            'raw_insert_count': 1,
        })

        # Double-check the types.
        self.assertIsInstance(filediff.extra_data['parent_source_filename'],
                              str)
        self.assertIsInstance(filediff.extra_data['parent_source_revision'],
                              str)

        original_file = get_original_file(filediff=filediff,
                                          request=None)
        self.assertEqual(original_file, b'Foo\nBar\n')
        self.assertSpyCalled(patch)

        patched_file = get_patched_file(source_data=original_file,
                                        filediff=filediff)
        self.assertEqual(patched_file, b'Foo\nBar\nBaz\n')
        self.assertEqual(len(patch.calls), 2)


class ValidateCommitFormTests(SpyAgency, TestCase):
    """Unit tests for ValidateCommitForm."""

    fixtures = ['test_scmtools']

    _PARENT_DIFF_DATA = (
        b'diff --git a/README b/README\n'
        b'new file mode 100644\n'
        b'index 0000000..94bdd3e\n'
        b'--- /dev/null\n'
        b'+++ b/README\n'
        b'@@ -0,0 +2 @@\n'
        b'+blah blah\n'
        b'+blah blah\n'
    )

    @classmethod
    def setUpClass(cls):
        super(ValidateCommitFormTests, cls).setUpClass()

        cls.request_factory = RequestFactory()

    def setUp(self):
        super(ValidateCommitFormTests, self).setUp()

        self.repository = self.create_repository(tool_name='Git')
        self.request = self.request_factory.get('/')
        self.diff = SimpleUploadedFile('diff',
                                       self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                       content_type='text/x-patch')

    def test_clean_already_validated(self):
        """Testing ValidateCommitForm.clean for a commit that has already been
        validated
        """
        validation_info = self._base64_json({
            'r1': {
                'parent_id': 'r0',
                'tree': {
                    'added': [],
                    'removed': [],
                    'modified': [],
                },
            },
        })

        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r1',
                'parent_id': 'r0',
                'validation_info': validation_info,
            },
            files={
                'diff': self.diff,
            })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {
            'validation_info': ['This commit was already validated.'],
        })

    def test_clean_parent_not_validated(self):
        """Testing ValidateCommitForm.clean for a commit whose parent has not
        been validated
        """
        validation_info = self._base64_json({
            'r1': {
                'parent_id': 'r0',
                'tree': {
                    'added': [],
                    'removed': [],
                    'modified': [],
                },
            },
        })

        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r3',
                'parent_id': 'r2',
                'validation_info': validation_info,
            },
            files={
                'diff': self.diff,
            })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {
            'validation_info': ['The parent commit was not validated.'],
        })

    def test_clean_parent_diff_subsequent_commit(self):
        """Testing ValidateCommitForm.clean with a non-empty parent diff for
        a subsequent commit
        """
        validation_info = self._base64_json({
            'r1': {
                'parent_id': 'r0',
                'tree': {
                    'added': [],
                    'removed': [],
                    'modified': [],
                },
            },
        })

        parent_diff = SimpleUploadedFile('diff',
                                         self._PARENT_DIFF_DATA,
                                         content_type='text/x-patch')

        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r2',
                'parent_id': 'r1',
                'validation_info': validation_info,
            },
            files={
                'diff': self.diff,
                'parent_diff': parent_diff,
            })

        self.assertTrue(form.is_valid())

    def test_clean_validation_info(self):
        """Testing ValidateCommitForm.clean_validation_info"""
        validation_info = self._base64_json({
            'r1': {
                'parent_id': 'r0',
                'tree': {
                    'added': [],
                    'removed': [],
                    'modified': [],
                },
            },
        })

        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r2',
                'parent_id': 'r1',
                'validation_info': validation_info,
            },
            files={
                'diff': self.diff,
            })

        self.assertTrue(form.is_valid())

    def test_clean_validation_info_invalid_base64(self):
        """Testing ValidateCommitForm.clean_validation_info with
        non-base64-encoded data"""
        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r2',
                'parent_id': 'r1',
                'validation_info': 'This is not base64!',
            },
            files={
                'diff': self.diff,
            })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {
            'validation_info': [
                'Could not parse validation info "This is not base64!": '
                'Incorrect padding',
            ],
        })

    def test_clean_validation_info_invalid_json(self):
        """Testing ValidateCommitForm.clean_validation_info with base64-encoded
        non-json data
        """
        validation_info = base64.b64encode(b'Not valid json.')
        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r2',
                'parent_id': 'r1',
                'validation_info': validation_info.decode('utf-8'),
            },
            files={
                'diff': self.diff,
            })

        self.assertFalse(form.is_valid())

        expected_error = 'Expecting value: line 1 column 1 (char 0)'

        self.assertEqual(form.errors, {
            'validation_info': [
                'Could not parse validation info "%s": %s'
                % (validation_info.decode('utf-8'), expected_error),
            ],
        })

    def test_validate_diff(self):
        """Testing ValidateCommitForm.validate_diff"""
        self.spy_on(self.repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)
        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r1',
                'parent_id': 'r2',
            },
            files={
                'diff': self.diff,
            })

        self.assertTrue(form.is_valid())
        form.validate_diff()

    def test_validate_diff_subsequent_commit(self):
        """Testing ValidateCommitForm.validate_diff for a subsequent commit"""
        diff_content = (
            b'diff --git a/foo b/foo\n'
            b'index %s..%s 100644\n'
            b'--- a/foo\n'
            b'+++ b/foo\n'
            b'@@ -0,0 +1,2 @@\n'
            b'+This is not a new file.\n'
            % (b'a' * 40, b'b' * 40)
        )
        diff = SimpleUploadedFile('diff', diff_content,
                                  content_type='text/x-patch')

        validation_info = self._base64_json({
            'r1': {
                'parent_id': 'r0',
                'tree': {
                    'added': [{
                        'filename': 'foo',
                        'revision': 'a' * 40,
                    }],
                    'removed': [],
                    'modified': [],
                },
            },
        })

        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r2',
                'parent_id': 'r1',
                'validation_info': validation_info,
            },
            files={
                'diff': diff,
            })

        self.assertTrue(form.is_valid())
        form.validate_diff()

    def test_validate_diff_missing_files(self):
        """Testing ValidateCommitForm.validate_diff for a subsequent commit
        with missing files
        """
        validation_info = self._base64_json({
            'r1': {
                'parent_id': 'r0',
                'tree': {
                    'added': [],
                    'removed': [],
                    'modified': [],
                },
            },
        })

        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r2',
                'parent_id': 'r1',
                'validation_info': validation_info,
            },
            files={
                'diff': self.diff,
            })

        self.assertTrue(form.is_valid())

        with self.assertRaises(FileNotFoundError):
            form.validate_diff()

    def test_validate_diff_empty(self):
        """Testing ValidateCommitForm.validate_diff for an empty diff"""
        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r1',
                'parent_id': 'r0',
            },
            files={
                'diff': SimpleUploadedFile('diff', b' ',
                                           content_type='text/x-patch'),
            })

        self.assertTrue(form.is_valid())

        with self.assertRaises(EmptyDiffError):
            form.validate_diff()

    def test_validate_diff_too_big(self):
        """Testing ValidateCommitForm.validate_diff for a diff that is too
        large
        """
        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r1',
                'parent_id': 'r0',
            },
            files={
                'diff': self.diff,
            })

        self.assertTrue(form.is_valid())

        with self.assertRaises(DiffTooBigError):
            with self.siteconfig_settings({'diffviewer_max_diff_size': 1},
                                          reload_settings=False):
                form.validate_diff()

    def test_validate_diff_parser_error(self):
        """Testing ValidateCommitForm.validate_diff for an invalid diff"""
        form = ValidateCommitForm(
            repository=self.repository,
            request=self.request,
            data={
                'commit_id': 'r1',
                'parent_id': 'r0',
            },
            files={
                'diff': SimpleUploadedFile('diff', b'asdf',
                                           content_type='text/x-patch'),
            })

        self.assertTrue(form.is_valid())

        with self.assertRaises(DiffParserError):
            form.validate_diff()

    def _base64_json(self, data):
        """Return a Base64-encoded JSON payload.

        Args:
            data (object):
                The data to encode to JSON.

        Returns:
            unicode:
            The Base64-encoded JSON payload.
        """
        content = json.dumps(data).encode('utf-8')
        return base64.b64encode(content).decode('utf-8')



import os

from django.core.management.base import BaseCommand

from reviewboard.reviews.models import Screenshot


class Command(BaseCommand):
    def handle(self, **options):
        prefix = os.path.join('images', 'uploaded')
        new_prefix = os.path.join('uploaded', 'images')

        for screenshot in Screenshot.objects.all():
            if screenshot.image.startswith(prefix):
                screenshot.image = \
                    os.path.join(new_prefix,
                                 os.path.basename(screenshot.image))
                screenshot.save()



import logging

from django.core.exceptions import MultipleObjectsReturned
from django.db.models import Manager
from djblets.db.managers import ConcurrencyManager

from reviewboard.accounts.trophies import trophies_registry


logger = logging.getLogger(__name__)


class LocalSiteProfileManager(ConcurrencyManager):
    """Manager for Local Site profiles."""

    def for_user(self, user, profile, local_site, create_if_missing=False):
        """Return the Local Site profile for the given user and site.

        Version Added:
            5.0

        Args:
            user (django.contrib.auth.models.User):
                The user to get the profile for.

            profile (reviewboard.accounts.models.Profile):
                The user's global profile.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site to get the profile for. This may be ``None``.

            create_if_missing (bool, optional):
                Whether to create the LocalSiteProfile if it is missing.

        Returns:
            tuple:
            A two-tuple, containing:

            1. The :py:class:`~reviewboard.accounts.models.LocalSiteProfile`
            2. A :py:class:`bool` indicating whether the profile is newly
               created.

        Raises:
            reviewboard.accounts.models.LocalSiteProfile.DoesNotExist:
                If ``create_if_missing`` is False and the profile does not
                exist, this can be raised.
        """
        try:
            if create_if_missing:
                site_profile, is_new = self.get_or_create(
                    user=user,
                    profile=profile,
                    local_site=local_site)
            else:
                # This may raise LocalSiteProfile.DoesNotExist.
                site_profile = self.get(user=user,
                                        profile=profile,
                                        local_site=local_site)
                is_new = False
        except MultipleObjectsReturned:
            site_profile = self._fix_duplicate_profiles(
                user, profile, local_site)
            is_new = False

        return site_profile, is_new

    def _fix_duplicate_profiles(self, user, profile, local_site):
        """Fix the case where we end up with duplicate Local Site profiles.

        Until Review Board 5.0, we were not using ConcurrencyManager for
        LocalSiteProfile. In addition, MySQL had an issue where the
        unique_together constraint only worked properly when the local_site
        relation was non-NULL, potentially resulting in multiple profiles.

        In the case that we discover this, we'll fix it up by merging any
        duplicates and deleting the extras.

        Args:
            user (django.contrib.auth.models.User):
                The user to get the profile for.

            profile (reviewboard.accounts.models.Profile):
                The user's global profile.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site to get the profile for. This may be ``None``.

        Returns:
            reviewboard.accounts.models.LocalSiteProfile:
            The new, merged profile object.
        """
        profiles = list(self.filter(user=user,
                                    profile=profile,
                                    local_site=local_site))

        master_profile = profiles[0]

        # If any of the profiles have permissions set, prefer that.
        for profile in profiles[1:]:
            if profile.permissions and not master_profile.permissions:
                master_profile.permissions = profile.permissions

            profile.delete()

        # Just reset counts to None so they'll be recalculated.
        master_profile.direct_incoming_request_count = None
        master_profile.total_incoming_request_count = None
        master_profile.pending_outgoing_request_count = None
        master_profile.total_outgoing_request_count = None
        master_profile.starred_public_request_count = None
        master_profile.save(update_fields=(
            'permissions',
            'direct_incoming_request_count',
            'total_incoming_request_count',
            'pending_outgoing_request_count',
            'total_outgoing_request_count',
            'starred_public_request_count',
        ))

        return master_profile


class ProfileManager(Manager):
    """Manager for user profiles."""

    def get_or_create(self, user, *args, **kwargs):
        """Return the profile for the user.

        This will create the profile if one does not exist.
        """
        if hasattr(user, '_profile'):
            return user._profile, False

        profile, is_new = \
            super(ProfileManager, self).get_or_create(user=user, *args,
                                                      **kwargs)
        user._profile = profile

        return profile, is_new


class ReviewRequestVisitManager(ConcurrencyManager):
    """Manager for review request visits."""

    def unarchive_all(self, review_request):
        """Unarchive a review request for all users.

        Unarchives the given review request for all users by changing all
        review request visit database entries for this review request from
        archived to visible.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request to unarchive.
        """
        queryset = self.filter(review_request=review_request,
                               visibility=self.model.ARCHIVED)
        queryset.update(visibility=self.model.VISIBLE)

    def update_visibility(self, review_request, user, new_visibility):
        """Update the visibility of a review request.

        This will set the review request as visible, archived, or muted for the
        given user.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request to update the visibility of.

            user (django.contrib.auth.models.User):
                The current user.

            new_visibility (unicode):
                The new visibility to update the review request to. This will
                be one of
                :py:attr:`~reviewboard.reviews.models.review_request.ReviewRequest.VISIBLE`,
                :py:attr:`~reviewboard.reviews.models.review_request.ReviewRequest.ARCHIVED`,
                or
                :py:attr:`~reviewboard.reviews.models.review_request.ReviewRequest.MUTED`.

        Returns:
            reviewboard.accounts.models.ReviewRequestVisit:
            The review request visit.
        """
        return self.update_or_create(
            user=user,
            review_request=review_request,
            defaults={
                'visibility': new_visibility,
            })[0]


class TrophyManager(Manager):
    """Manager for trophies.

    Creates new trophies, updates the database and fetches trophies from the
    database.
    """

    def compute_trophies(self, review_request):
        """Compute and return trophies for a review request.

        Computes trophies for a given review request by looping through all
        registered trophy types and seeing if any apply to the review request.

        If trophies are to be awarded, they are saved in the database and
        returned. If no trophies are to be awarded, an empty list is returned.
        """
        if 'calculated_trophies' in review_request.extra_data:
            return list(self.filter(review_request=review_request))

        calculated_trophy_types = []

        for registered_trophy_type in trophies_registry:
            try:
                instance = registered_trophy_type()
            except Exception as e:
                logger.error('Error instantiating trophy type %r: %s',
                             registered_trophy_type, e, exc_info=True)
                continue

            try:
                if instance.qualifies(review_request):
                    calculated_trophy_types.append(instance)
            except Exception as e:
                logger.error('Error when running %r.instance_qualifies: %s',
                             registered_trophy_type, e, exc_info=True)

        trophies = [
            self.model.objects.create(category=trophy_type.category,
                                      review_request=review_request,
                                      local_site=review_request.local_site,
                                      user=review_request.submitter)
            for trophy_type in calculated_trophy_types
        ]

        review_request.extra_data['calculated_trophies'] = True
        review_request.save(update_fields=['extra_data'])

        return trophies

    def get_trophies(self, review_request):
        """Get all the trophies for a given review request."""
        return self.compute_trophies(review_request)



"""Unit tests for reviewboard.extensions.hooks.FileDiffACLHook."""

import kgb
from djblets.features.testing import override_feature_check

from reviewboard.extensions.hooks import FileDiffACLHook
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase
from reviewboard.reviews.features import DiffACLsFeature


class FileDiffACLHookTests(kgb.SpyAgency, BaseExtensionHookTestCase):
    """Tests for the FileDiffACLHook."""

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self):
        super(FileDiffACLHookTests, self).setUp()

        self.user = self.create_user()
        self.review_request = self.create_review_request(
            create_repository=True)
        self.review_request.target_people.add(self.review_request.submitter)
        self.create_diffset(review_request=self.review_request, draft=True)
        self.review_request.publish(user=self.review_request.submitter)

    def test_single_aclhook_true(self):
        """Testing FileDiffACLHook basic approval with True result"""
        self._test_hook_approval_sequence([True], True)

    def test_single_aclhook_none(self):
        """Testing FileDiffACLHook basic approval with None result"""
        self._test_hook_approval_sequence([None], True)

    def test_single_aclhook_false(self):
        """Testing FileDiffACLHook basic approval with False result"""
        self._test_hook_approval_sequence([False], False)

    def test_multiple_aclhooks_1(self):
        """Testing FileDiffACLHook multiple with True and False"""
        self._test_hook_approval_sequence([True, False], False)

    def test_multiple_aclhooks_2(self):
        """Testing FileDiffACLHook multiple with True and None"""
        self._test_hook_approval_sequence([True, None], True)

    def test_multiple_aclhooks_3(self):
        """Testing FileDiffACLHook multiple with False and None"""
        self._test_hook_approval_sequence([False, None], False)

    def _test_hook_approval_sequence(self, accessible_values, result):
        """Test a sequence of FileDiffACLHook approval results.

        Args:
            accessible_values (list of bool):
                A list of the values to return from FileDiffACLHook
                implementations.

            result (bool):
                A resulting approval value to check.
        """
        with override_feature_check(DiffACLsFeature.feature_id,
                                    enabled=True):
            for value in accessible_values:
                hook = FileDiffACLHook(extension=self.extension)
                self.spy_on(hook.is_accessible, op=kgb.SpyOpReturn(value))

            self.assertEqual(self.review_request.is_accessible_by(self.user),
                             result)



"""Unit tests for the groups page.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Optional, Sequence, TYPE_CHECKING

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import User
from django.db.models import Count, Q
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.models import Profile
from reviewboard.datagrids.tests.base import BaseViewTestCase
from reviewboard.reviews.models import Group
from reviewboard.reviews.testing.queries.review_groups import (
    get_review_groups_accessible_prep_equeries,
    get_review_groups_accessible_q,
)
from reviewboard.site.models import LocalSite
from reviewboard.testing.queries.http import get_http_request_start_equeries

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries


class GroupListViewTests(BaseViewTestCase):
    """Unit tests for the group_list view."""

    datagrid_url = '/groups/'

    @add_fixtures(['test_users'])
    def test_with_access(self):
        """Testing group_list view"""
        self._test_with_access()

    @add_fixtures(['test_site', 'test_users'])
    def test_with_access_with_local_site(self):
        """Testing group_list view on a Local Site"""
        self._test_with_access(with_local_site=True,
                               local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_with_access_with_local_sites_in_db(self):
        """Testing group_list view with Local Sites in the database"""
        self._test_with_access(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_as_anonymous_and_redirect(self):
        """Testing group_list view with site-wide login enabled"""
        with self.siteconfig_settings({'auth_require_sitewide_login': True},
                                      reload_settings=False):
            response = self.client.get('/groups/')

        self.assertEqual(response.status_code, 302)

    def _test_with_access(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for a user with access to the datagrid.

        This accesses the groups datagrid as a logged-in user with access
        to the global site or Local Site (depending on the test), checking
        queries and datagrid row results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        review_groups = [
            self.create_review_group(local_site=local_site,
                                     name='group-%02d' % (i + 1))
            for i in range(10)
        ]

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            review_groups=review_groups)

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site))

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(
            [
                row['object']
                for row in datagrid.rows
            ],
            review_groups)

    def _build_datagrid_equeries(
        self,
        *,
        user: User,
        profile: Profile,
        local_site: Optional[LocalSite] = None,
        local_sites_in_db: bool = False,
        review_groups: Sequence[Group],
    ) -> ExpectedQueries:
        """Return expected queries for viewing the datagrid.

        This assumes that the user has access to the datagrid, and that any
        cacheable state is already cached.

        Version Added:
            5.0.7

        Args:
            user (django.contrib.auth.models.User):
                The user accessing the datagrid.

            profile (reviewboard.accounts.models.Profile):
                The user's profile.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site that's being accessed, if any.

            local_sites_in_db (bool, optional):
                Whether the database contains any Local Sites.

            review_groups (list of reviewboard.reviews.models.group.Group):
                The list of review groups that should be returned in the
                results, in display order.

        Returns:
            list of dict:
            The list of expected queries.
        """
        review_group_pks = [
            _group.pk
            for _group in review_groups
        ]

        if local_site is not None:
            row_data_select_related = {
                'local_site',
            }
        else:
            row_data_select_related = set()

        rows_q_result = get_review_groups_accessible_q(
            user=user,
            local_site=local_site)
        rows_q = rows_q_result['q']
        rows_q_tables = rows_q_result['tables']
        rows_q_num_joins = len(rows_q_tables) - 1

        equeries = get_http_request_start_equeries(
            user=user,
            local_site=local_site)
        equeries += get_review_groups_accessible_prep_equeries(
            user=user,
            local_site=local_site,
            needs_local_site_profile_query=True)
        equeries += [
            {
                '__note__': 'Update datagrid state on the user profile',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=profile.pk),
            },
            {
                '__note__': (
                    'Fetch the number of items across all datagrid pages'
                ),
                'annotations': {'__count': Count('*')},
                'model': Group,

                'subqueries': [
                    {
                        'distinct': True,
                        'join_types': {
                            'reviews_group_users': 'LEFT OUTER JOIN',
                        },
                        'model': Group,
                        'num_joins': rows_q_num_joins,
                        'subquery': True,
                        'tables': rows_q_tables,
                        'where': rows_q,
                    },
                ],
            },
            {
                '__note__': 'Fetch the IDs of the items for one page',
                'distinct': True,
                'join_types': {
                    'reviews_group_users': 'LEFT OUTER JOIN',
                },
                'limit': 10,
                'model': Group,
                'num_joins': rows_q_num_joins,
                'order_by': ('name',),
                'tables': rows_q_tables,
                'values_select': ('pk',),
                'where': rows_q,
            },
            {
                '__note__': (
                    "Fetch the IDs of the page's groups that are starred."
                ),
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'values_select': ('pk',),
                'where': (Q(starred_by__id=profile.pk) &
                          Q(pk__in=review_group_pks)),
            },
            {
                '__note__': (
                    'Fetch the data for one page based on the IDs.'
                ),
                'annotations': {
                    'column_pending_review_request_count': Count(
                        'review_requests',
                        filter=(Q(review_requests__public=True) &
                                Q(review_requests__status='P'))),
                },
                'group_by': True,
                'join_types': {
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                },
                'model': Group,
                'num_joins': 2,
                'select_related': row_data_select_related,
                'tables': {
                    'reviews_group',
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_target_groups',
                },
                'where': Q(pk__in=review_group_pks),
            },
        ]

        return equeries



"""Add a Repository.visible field.

Version Added:
    1.5
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('Repository', 'visible', models.BooleanField, initial=True)
]



"""Add ReviewRequestVisit and LocalSiteProfile.unique_together state.

This is needed for Django Evolution 0.7.

Version Added:
    2.0
"""

from django_evolution.mutations import ChangeMeta


MUTATIONS = [
    ChangeMeta('ReviewRequestVisit', 'unique_together',
               [('user', 'review_request')]),
    ChangeMeta('LocalSiteProfile', 'unique_together',
               [('user', 'local_site'), ('profile', 'local_site')]),
]



#!/usr/bin/env python
"""Prepare a Review Board tree for development."""

from __future__ import annotations

import argparse
import os
import platform
import re
import stat
import subprocess
import sys
from configparser import ConfigParser
from importlib import import_module
from random import choice
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from reviewboard.cmdline.rbsite import Site


FAQ_URL = \
    'https://notion.so/reviewboard/FAQ-6a19618dd534476ea844cba9a3576868'


# Our git post-checkout hook script which ensures stale .pyc files are not left
# around when switching branches, especially when switching between releases.
_POST_CHECKOUT = (
    "#!/bin/sh\n"
    "find . -iname '*.pyc' -delete\n"
)


#: The console instance to use for all output.
#:
#: Type:
#:     reviewboard.cmdline.utils.console.Console
console = None


class SiteOptions(object):
    """The site options."""

    copy_media = platform.system() == "Windows"


def create_settings(options):
    """Create a settings_local.py file if it doesn't exist.

    Args:
        options (argparse.Namespace):
            The options parsed from :py:mod:`argparse`.
    """
    if not os.path.exists('settings_local.py'):
        console.header('Creating your settings_local.py')

        console.print(
            'This settings_local.py will be placed in the current directory. '
            'It can be modified to point to a different database, or to '
            'enable custom Django settings.')

        # TODO: Use an actual templating system.
        src_path = os.path.join('contrib', 'conf', 'settings_local.py.tmpl')

        # XXX: Once we switch to Python 2.7+, use the multiple form of 'with'
        in_fp = open(src_path, 'r')
        out_fp = open('settings_local.py', 'w')

        for line in in_fp:
            if line.startswith('SECRET_KEY = '):
                secret_key = ''.join([
                    choice(
                        'abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)')
                    for i in range(50)
                ])

                out_fp.write('SECRET_KEY = "%s"\n' % secret_key)
            elif line.strip().startswith("'ENGINE': "):
                out_fp.write("        'ENGINE': 'django.db.backends.%s',\n" %
                             options.db_type)
            elif line.strip().startswith("'NAME': "):
                if options.db_type == 'sqlite3':
                    if options.db_name is not None:
                        name = os.path.abspath(options.db_name)
                        out_fp.write("        'NAME': '%s',\n" % name)
                    else:
                        out_fp.write("        'NAME': os.path.join(ROOT_PATH,"
                                     " 'reviewboard-%d.%d.db' % (VERSION[0], "
                                     "VERSION[1])),\n")
                else:
                    name = options.db_name
                    out_fp.write("        'NAME': '%s',\n" % name)
            elif line.strip().startswith("'USER': "):
                out_fp.write("        'USER': '%s',\n" % options.db_user)
            elif line.strip().startswith("'PASSWORD': "):
                out_fp.write("        'PASSWORD': '%s',\n"
                             % options.db_password)
            else:
                out_fp.write(line)

        in_fp.close()
        out_fp.close()


def install_git_hooks():
    """Install a post-checkout hook to delete pyc files."""
    console.header('Setting up your Git tree')
    console.print(
        'Your Git tree will be set up with a default post-checkout hook '
        'that clears any compiled Python files when switching branches.')

    try:
        gitdir = (
            subprocess.check_output(['git', 'rev-parse', '--git-common-dir'])
            .decode('utf-8')
            .strip()
        )
    except subprocess.CalledProcessError:
        console.error('Could not determine git directory. Are you in a '
                      'checkout?')
        return

    hook_path = os.path.join(gitdir, 'hooks', 'post-checkout')

    exc_mask = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH

    try:
        statinfo = os.stat(hook_path)
    except OSError:
        # The file does not exist.
        statinfo = None

    if statinfo:
        # The file exists. We need to determine if we should write to it.

        if statinfo.st_size != 0 and statinfo.st_mode & exc_mask:
            # The file is non-empty and executable, which means this *isn't*
            # the default hook that git installs when you create a new
            # repository.
            #
            # Let's check the hook's contents to see if its a hook we installed
            # previously or if the user has already set up their own hook.
            with open(hook_path, 'r') as f:
                contents = f.read()

            if contents != _POST_CHECKOUT:
                console.error(
                    'The hook "%(hook_path)s" already exists and differs '
                    'from the hook we would install. The existing hook '
                    'will be left alone.'
                    '\n'
                    'If you want this hook installed, please add the '
                    'following to that file:'
                    '\n'
                    '%(script)s'
                    % {
                        'hook_path': hook_path,
                        'script': '\n'.join(_POST_CHECKOUT.split('\n')[1:]),
                    })

                return

    # At this point we know we are safe to write to the hook file. This is
    # because one of the following is true:
    #
    # 1. The hook file does not exist.
    # 2. The hook file exists but is empty.
    # 3. The hook file exists but is non-executable (i.e., it contains a
    #    sample git hook).
    with open(hook_path, 'w') as f:
        f.write(_POST_CHECKOUT)
        os.fchmod(f.fileno(), 0o740)

    console.print('The post-checkout hook has been installed.')


def install_media(
    site: Site,
) -> None:
    """Install static media.

    Args:
        site (reviewboard.cmdline.rbsite.Site):
            The site to install media for.

            This will be the site corresponding to the current working
            directory.
    """
    from pipeline.collector import default_collector
    from pipeline.packager import Packager

    console.header('Setting up static media files')

    media_path = os.path.join('htdocs', 'media')
    uploaded_path = os.path.join(site.install_dir, media_path, 'uploaded')
    ext_media_path = os.path.join(site.install_dir, media_path, 'ext')

    site.mkdir(uploaded_path)
    site.mkdir(os.path.join(uploaded_path, 'images'))
    site.mkdir(ext_media_path)

    # Run Pipeline on all the files, so we can prime the static media
    # directory. This cuts down on the very long initial load times, in
    # exchange for a somewhat long up-front time. Pipeline will at least
    # compile files within each bundle in parallel.
    default_collector.collect()

    packager = Packager()

    package_types = (
        ('css', 'CSS'),
        ('js', 'JavaScript'),
    )

    total_packages = sum(
        len(packager.packages[package_type])
        for package_type, package_type_desc in package_types
    )

    i = 1

    for package_type, package_type_desc in package_types:
        packages = packager.packages[package_type]

        for package_name, package in sorted(packages.items(),
                                            key=lambda pair: pair[0]):
            console.progress_step(
                'Compiling %s bundle %s' % (package_type_desc,
                                            package.output_filename),
                step_num=i,
                total_steps=total_packages,
                func=lambda: packager.compile(package.paths))

            i += 1


def setup_local_packages() -> None:
    """Set up links to any standard local development packages.

    If there's a local editable build of Djblets available, it will be used
    for the editable install for Review Board.

    Version Added:
        8.0
    """
    packages: dict[str, str] = {}

    for package_name, module_name in (('Djblets', 'djblets'),
                                      ('django-pipeline', 'pipeline')):
        try:
            mod = import_module(module_name)
        except ImportError:
            # This is not installed, so we can skip it.
            continue

        tree_path = os.path.abspath(os.path.join(mod.__path__[0], '..'))

        if (os.path.basename(tree_path) != 'site-packages' and
            (os.path.exists(os.path.join(tree_path, 'setup.py')) or
             os.path.exists(os.path.join(tree_path, 'pyproject.toml')))):
            # This looks like a local editable install.
            packages[package_name] = tree_path

    if packages:
        local_packages_dir = '.local-packages'

        os.makedirs(local_packages_dir, 0o755, exist_ok=True)

        for package_name, tree_path in packages.items():
            local_package_link = os.path.join(local_packages_dir,
                                              package_name)

            if os.path.lexists(local_package_link):
                os.unlink(local_package_link)

            os.symlink(tree_path, local_package_link,
                       target_is_directory=True)


def install_dependencies(
    options: argparse.Namespace,
) -> None:
    """Install Review Board and all dependencies in editable mode.

    Version Changed:
        8.0:
        This now uses :command:`pip install -e`, and supports multiple
        Python versions using virtualenv-multiver's :command:`pydo`.

    Args:
        options (argparse.Namespace):
            The parsed command line arguments.
    """
    python_exes: list[str] = []

    # We can't use console.print() or console.header() here, since we're
    # running before we know we even have Django installed.
    print('Bootstrapping: Installing the Review Board package and '
          'dependencies..')

    if options.all_pyvers:
        # We need to determine which Python versions to use.
        #
        # Back when we were setup.py-based, we just had a hard-coded list of
        # versions to try, and we'd invoke a flag that would loop through the
        # versions in setup.py. Now, that looping logic is here, and it's
        # smarter.
        #
        # We grab the list of supported versions from pyproject.toml, and we
        # AND that with the list of versions supported by the virtualenv
        # (which requires a virtualenv-multiver-based approach).
        virtualenv_path = os.environ.get('VIRTUAL_ENV')

        if virtualenv_path:
            pydorc_path = os.path.join(virtualenv_path, '.pydorc')
        else:
            pydorc_path = None

        if not pydorc_path or not os.path.exists(pydorc_path):
            sys.stderr.write(
                '--all-pyvers can only be used when run in a '
                'virtualenv built using virtualenv-multiver.\n')
            sys.exit(1)

        # Read .pydorc.
        config_parser = ConfigParser()
        config_parser.read(pydorc_path)
        pydorc_pyvers = set(config_parser.get('pydo', 'pyvers').split(' '))

        # Read pyproject.toml.
        pyproject_pyvers: set[str] = set()

        with open('pyproject.toml', 'r', encoding='utf-8') as fp:
            PYVER_RE = re.compile(
                r"'Programming Language :: Python :: (\d\.\d+)'"
            )

            for line in fp:
                m = PYVER_RE.search(line)

                if m:
                    pyproject_pyvers.add(m.group(1))

        # Generate a list of all the Python executables we'll run.
        python_exes = [
            f'python{pyver}'
            for pyver in sorted(
                pyproject_pyvers & pydorc_pyvers,
                key=lambda ver: [
                    int(part)
                    for part in ver.split('.')
                ])
        ]
    else:
        python_exes = [sys.executable]

    for python_exe in python_exes:
        os.system(subprocess.list2cmdline([
            python_exe, '-m', 'pip', 'install', '-e', '.',
        ]))

    # `pip install -e` will try to manage .npm-workspaces, but it doesn't
    # know the packages in the containing virtualenv. If we didn't build
    # using a local editable copy of Djblets, then we're going to have a
    # dangling symlink. We need to fix this up here.
    try:
        import djblets
    except ImportError:
        sys.stderr.write('Could not import djblets. Something must have '
                         'gone wrong during environment setup.\n')
        sys.exit(1)

    djblets_link_path = os.path.join('.npm-workspaces', 'djblets')

    if os.path.lexists(djblets_link_path):
        os.unlink(djblets_link_path)

    os.symlink(djblets.__path__[0], djblets_link_path)


def create_superuser(site):
    """Create an initial superuser for the site.

    This will ask for a username, password, and e-mail address for the
    initial superuser account.

    If a superuser already exists (due to re-running this script on an
    existing database), it will be displayed for reference, and the user
    will be instructed on how to create a new one.

    Args:
        site (reviewboard.cmdline.rbsite.Site):
            The site to create the superuser on.
    """
    from django.contrib.auth.management import get_default_username
    from django.contrib.auth.models import User

    console.header('Set up an administrator account')

    admins = list(
        User.objects.filter(is_superuser=True)
        .values_list('username', flat=True)
    )

    if admins:
        console.print(
            'Existing admin account(s) were found: %(admins)s'
            '\n'
            'To create a new one, run:'
            % {
                'admins': ', '.join(admins),
            })
        console.print()
        console.print(
            '    ./reviewboard/manage.py createsuperuser',
            wrap=False)
    else:
        console.print("Now you'll need to set up a superuser (an admin "
                      "account). This will be used to log in and configure "
                      "Review Board.")

        site.admin_user = console.prompt_input(
            'Username',
            default=get_default_username() or 'admin')
        site.admin_password = console.prompt_input(
            'Password',
            prompt_type=console.PROMPT_TYPE_PASSWORD)

        while True:
            confirmed_password = console.prompt_input(
                'Confirm Password',
                prompt_type=console.PROMPT_TYPE_PASSWORD)

            if confirmed_password == site.admin_password:
                break

            console.error('Passwords must match.')

        site.admin_email = console.prompt_input('E-Mail Address')

        site.create_admin_user()


def parse_options(args):
    """Parse the command-line arguments and return the results.

    Args:
        args (list of bytes):
            The arguments to parse.

    Returns:
        argparse.Namespace:
        The parsed arguments.
    """
    parser = argparse.ArgumentParser(
        'Prepare a Review Board tree for development.',
        usage='%(prog)s [options]')

    parser.add_argument(
        '--all-pyvers',
        action='store_true',
        default=False,
        help=(
            'Set up the package using all supported versions of Python. '
            'This requires a multi-Python virtualenv supporting all the '
            'versions required.'
        ))

    parser.add_argument(
        '--only',
        action='store_true',
        dest='only',
        help=(
            'Disable all feature by default (implying --no-<feature> for each '
            'feature) unless explicitly specified (e.g., "--only --media '
            '--deps" to only install media and dependencies).'
        ))

    parser.add_argument(
        '--media',
        action='store_true',
        dest='install_media',
        default=None,
        help='Install media files when --only specified')
    parser.add_argument(
        '--no-media',
        action='store_false',
        dest='install_media',
        help="Don't install media files")

    parser.add_argument(
        '--db',
        action='store_true',
        dest='sync_db',
        default=None,
        help='Synchronize database when --only specified')
    parser.add_argument(
        '--no-db',
        action='store_false',
        dest='sync_db',
        help="Don't synchronize the database")

    parser.add_argument(
        '--deps',
        action='store_true',
        dest='install_deps',
        default=None,
        help='Install dependencies when --only specified')
    parser.add_argument(
        '--no-deps',
        action='store_false',
        dest='install_deps',
        help="Don't install dependencies")

    parser.add_argument(
        '--hooks',
        action='store_true',
        dest='install_hooks',
        default=None,
        help='Install repository hooks when --only specified')
    parser.add_argument(
        '--no-hooks',
        action='store_false',
        dest='install_hooks',
        help="Don't install repository hooks")

    parser.add_argument(
        '--database-type',
        dest='db_type',
        default='sqlite3',
        help="Database type (postgresql, mysql, sqlite3)")
    parser.add_argument(
        '--database-name',
        dest='db_name',
        default=None,
        help="Database name (or path, for sqlite3)")
    parser.add_argument(
        '--database-user',
        dest='db_user',
        default='',
        help="Database user")
    parser.add_argument(
        '--database-password',
        dest='db_password',
        default='',
        help="Database password")

    options = parser.parse_args(args)

    # Post-process the options so that anything that didn't get explicitly set
    # will have a boolean value.
    for attr in ('install_media', 'install_hooks', 'install_deps', 'sync_db'):
        if getattr(options, attr) is None:
            setattr(options, attr, not options.only)

    return options


def main():
    """The entry point of the prepare-dev script."""
    global console

    if not os.path.exists(os.path.join("reviewboard", "manage.py")):
        sys.stderr.write("This must be run from the top-level Review Board "
                         "directory\n")
        sys.exit(1)

    options = parse_options(sys.argv[1:])

    setup_local_packages()

    if options.install_deps:
        install_dependencies(options)

    # Insert the current directory first in the module path so we find the
    # correct reviewboard package.
    sys.path.insert(0, os.getcwd())

    from reviewboard.cmdline import rbsite
    from reviewboard.cmdline.rbsite import Site, setup_rbsite
    from reviewboard.cmdline.utils.console import get_console

    setup_rbsite()

    console = get_console()
    console.allow_color = True

    console.header('Welcome to Review Board!',
                   leading_newlines=False)
    console.print(
        "Let's get your development environment set up and ready to go. This "
        "will set up your settings_local.py file, your database, initial "
        "static media files, and prepare an administrator account."
        "\n"
        "If you have any issues, first see if it's answered in our FAQ:"
        "\n"
        "%(faq_url)s"
        "\n"
        "If you're a student working on Review Board, you can also get help "
        "from your mentors. If you're a contributor, please contact:"
        "\n"
        "reviewboard-dev@googlegroups.com"
        % {
            'faq_url': FAQ_URL,
        })

    # Re-use the Site class, since it has some useful functions.
    site_path = os.path.abspath('reviewboard')
    site = Site(site_path, SiteOptions)

    create_settings(options)

    if options.install_hooks:
        install_git_hooks()

    try:
        if options.sync_db:
            site.abs_install_dir = os.getcwd()

            console.header('Setting up the Review Board database')
            site.setup_settings()
            site.update_database(allow_input=True,
                                 report_progress=True)
    except KeyboardInterrupt:
        console.error(
            'The process was canceled in the middle of creating the database, '
            'which can result in a corrupted setup. Please remove the '
            'database file and run:')
        console.error(
            '    ./reviewboard/manage.py createdb',
            wrap=False)
        return

    if options.install_media:
        install_media(site)

    create_superuser(site)

    console.header('Your Review Board tree is ready for development!')
    console.print('You can now run your development server by running:')
    console.print()
    console.print('    ./contrib/internal/devserver.py',
                  wrap=False)
    console.print()


if __name__ == "__main__":
    main()



"""Built-in actions for the accounts app.

Version Added:
    6.0
"""

from typing import TYPE_CHECKING

from django.conf import settings
from django.template import Context
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from reviewboard import get_manual_url
from reviewboard.actions import (ActionPlacement,
                                 AttachmentPoint,
                                 BaseAction,
                                 BaseMenuAction)
from reviewboard.actions.renderers import MenuActionGroupRenderer


if TYPE_CHECKING:
    MixinParent = BaseAction
else:
    MixinParent = object


class LoggedInUserMixin(MixinParent):
    """Mixin for actions that only render for logged-in users.

    Version Added:
        6.0
    """

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        return (super().should_render(context=context) and
                request.user.is_authenticated)


class AccountMenuActionRenderer(MenuActionGroupRenderer):
    """Action renderer for the My Account menu.

    This provides a custom template used to render the menu for display
    in the page header, using the username and avatar.

    Version Added:
        8.0
    """

    template_name = 'accounts/account_menu_action.html'


class AccountMenuAction(LoggedInUserMixin, BaseMenuAction):
    """A menu for account-related actions.

    Version Added:
        6.0
    """

    action_id = 'account-menu'
    default_renderer_cls = AccountMenuActionRenderer
    label = ''

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER),
    ]


class LoginAction(BaseAction):
    """Action for logging in.

    Version Added:
        6.0
    """

    action_id = 'login'
    label = _('Log in')

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER),
    ]

    def get_url(
        self,
        *,
        context: Context,
    ) -> str:
        """Return the URL for the action.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            str:
            The URL to use for the action.
        """
        request = context['request']
        return '%s?next=%s' % (reverse('login'), request.path)

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        return (super().should_render(context=context) and
                not request.user.is_authenticated)


class LogoutAction(LoggedInUserMixin, BaseAction):
    """Action for logging out.

    Version Added:
        6.0
    """

    action_id = 'logout'
    label = _('Log out')
    url_name = 'logout'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=AccountMenuAction.action_id),
    ]


class AdminAction(BaseAction):
    """Action for the "Admin" page.

    Version Added:
        6.0
    """

    action_id = 'admin'
    label = _('Admin')
    url_name = 'admin-dashboard'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=AccountMenuAction.action_id),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']
        return (super().should_render(context=context) and
                request.user.is_staff)


class MyAccountAction(LoggedInUserMixin, BaseAction):
    """Action for the "My Account" page.

    Version Added:
        6.0
    """

    action_id = 'my-account'
    label = _('My account')
    url_name = 'user-preferences'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=AccountMenuAction.action_id),
    ]


class SupportMenuAction(BaseMenuAction):
    """A menu for support options.

    Version Added:
        6.0
    """

    action_id = 'support-menu'
    label = _('Support')

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER),
    ]


class DocumentationAction(BaseAction):
    """Action for accessing Review Board documentation.

    Version Added:
        6.0
    """

    action_id = 'documentation'
    label = _('Documentation')

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=SupportMenuAction.action_id),
    ]

    def get_url(
        self,
        *,
        context: Context,
    ) -> str:
        """Return the URL for the action.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            str:
            The URL to use for the action.
        """
        return get_manual_url()


class SupportAction(BaseAction):
    """Action for linking to support.

    Version Added:
        6.0
    """

    action_id = 'support'
    label = _('Get Support')
    url_name = 'support'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=SupportMenuAction.action_id),
    ]


class FollowMenuAction(BaseMenuAction):
    """A menu for follow options.

    Version Added:
        6.0
    """

    action_id = 'follow-menu'
    label = _('Follow')

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER),
    ]

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        return (super().should_render(context=context) and
                not getattr(settings, 'DISABLE_FOLLOW_MENU', False))


class FollowNewsAction(BaseAction):
    """Action for following via News.

    Version Added:
        6.0
    """

    action_id = 'follow-rss'
    label = _('Review Board News')
    icon_class = 'rb-icon-rss'
    url = 'https://www.reviewboard.org/news/'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=FollowMenuAction.action_id),
    ]


class FollowBlueSkyAction(BaseAction):
    """Action for following via BlueSky.

    Version Added:
        7.0
    """

    action_id = 'follow-bluesky'
    label = _('BlueSky')
    icon_class = 'rb-icon-brand-bluesky'
    url = 'https://bsky.app/profile/reviewboard.bsky.social'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=FollowMenuAction.action_id),
    ]


class FollowLinkedInAction(BaseAction):
    """Action for following via LinkedIn.

    Version Added:
        7.0
    """

    action_id = 'follow-linkedin'
    label = _('LinkedIn')
    icon_class = 'rb-icon-brand-linkedin'
    url = 'https://www.linkedin.com/company/reviewboard/'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=FollowMenuAction.action_id),
    ]


class FollowMastodonAction(BaseAction):
    """Action for following via Mastodon.

    Version Added:
        7.0
    """

    action_id = 'follow-mastodon'
    label = _('Mastodon')
    icon_class = 'rb-icon-brand-mastodon'
    url = 'https://mastodon.online/@reviewboard'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=FollowMenuAction.action_id),
    ]


class FollowTwitterAction(BaseAction):
    """Action for following via Twitter.

    Version Added:
        6.0
    """

    action_id = 'follow-twitter'
    label = _('Twitter')
    icon_class = 'rb-icon-brand-twitter'
    url = 'https://twitter.com/reviewboard/'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=FollowMenuAction.action_id),
    ]


class FollowFacebookAction(BaseAction):
    """Action for following via Facebook.

    Version Added:
        6.0
    """

    action_id = 'follow-facebook'
    label = _('Facebook')
    icon_class = 'rb-icon-brand-facebook'
    url = 'https://facebook.com/reviewboard.org'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=FollowMenuAction.action_id),
    ]


class FollowRedditAction(BaseAction):
    """Action for following via Reddit.

    Version Added:
        6.0
    """

    action_id = 'follow-reddit'
    label = _('Reddit')
    icon_class = 'rb-icon-brand-reddit'
    url = 'https://reddit.com/r/reviewboard'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=FollowMenuAction.action_id),
    ]


class FollowYouTubeAction(BaseAction):
    """Action for following via YouTube.

    Version Added:
        6.0
    """

    action_id = 'follow-youtube'
    label = _('YouTube')
    icon_class = 'rb-icon-brand-youtube'
    url = 'https://www.youtube.com/channel/UCTnwzlRTtx8wQOmyXiA_iCg'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=FollowMenuAction.action_id),
    ]



"""Unit tests for reviewboard.admin.form_widgets.LexersMappingWidget."""

from django import forms
from django.utils.html import escape, format_html
from pygments.lexers import get_all_lexers

from reviewboard.admin.form_widgets import LexersMappingWidget
from reviewboard.testing.testcase import TestCase


class TestForm(forms.Form):
    """A Test Form with a field that contains a LexersMappingWidget."""

    my_mapping_field = forms.CharField(
        label=('Lexer Mapping'),
        required=False,
        widget=LexersMappingWidget())


class LexersMappingWidgetTests(TestCase):
    """Unit tests for LexersMappingWidget."""

    def test_render_empty(self):
        """Testing LexersMappingWidget.render with no initial data"""
        my_form = TestForm()
        html = my_form.fields['my_mapping_field'].widget.render(
            'Lexer Mapping',
            (),
            {'id': 'lexer-mapping'})

        correct_html_parts = [
            '<input type="text" name="Lexer Mapping_0" id="lexer-mapping_0">',
            '<select name="Lexer Mapping_1" id="lexer-mapping_1">'
        ]

        for lex in get_all_lexers():
            lex_name = escape(lex[0])
            correct_html_parts.append(format_html(
                '<option value="{}">{}</option>', lex_name, lex_name))

        correct_html_parts.append('</select>')
        correct_html = ''.join(correct_html_parts)

        self.assertHTMLEqual(correct_html, html)

    def test_render_with_data(self):
        """Testing LexersMappingWidget.render with initial data"""
        my_form = TestForm()
        html = my_form.fields['my_mapping_field'].widget.render(
            'Lexer Mapping',
            ('.py', 'Python'),
            {'id': 'lexer-mapping'})

        correct_html_parts = [
            '<input type="text" name="Lexer Mapping_0"',
            'value=".py" id="lexer-mapping_0">',
            '<select name="Lexer Mapping_1" id="lexer-mapping_1">'
        ]

        for lex in get_all_lexers():
            lex_name = escape(lex[0])

            if lex_name == 'Python':
                correct_html_parts.append(format_html(
                    '<option value="{}" selected>{}</option>', lex_name,
                    lex_name))
            else:
                correct_html_parts.append(format_html(
                    '<option value="{}">{}</option>', lex_name, lex_name))

        correct_html_parts.append('</select>')
        correct_html = ''.join(correct_html_parts)

        self.assertHTMLEqual(correct_html, html)

    def test_value_from_datadict(self):
        """Testing LexersMappingWidget.value_from_datadict"""
        my_form = TestForm()
        value = (
            my_form.fields['my_mapping_field']
            .widget
            .value_from_datadict(
                {'mapping_0': '.py',
                 'mapping_1': 'Python'},
                {},
                'mapping'))
        self.assertEqual(value, ('.py', 'Python'))

    def test_value_from_datadict_with_no_data(self):
        """Testing LexersMappingWidget.value_from_datadict with no data"""
        my_form = TestForm()
        value = (
            my_form.fields['my_mapping_field']
            .widget
            .value_from_datadict(
                {'mapping_0': '',
                 'mapping_1': ''},
                {},
                'mapping'))
        self.assertEqual(value, ('', ''))

    def test_value_from_datadict_with_missing_data(self):
        """Testing LexersMappingWidget.value_from_datadict with missing data"""
        my_form = TestForm()
        value = (
            my_form.fields['my_mapping_field']
            .widget
            .value_from_datadict(
                {},
                {},
                'mapping'))
        self.assertEqual(value, (None, None))

    def test_decompress(self):
        """Testing LexersMappingWidget.decompress"""
        my_form = TestForm()
        value = (
            my_form.fields['my_mapping_field']
            .widget
            .decompress(('.py', 'Python')))
        self.assertEqual(value, ['.py', 'Python'])

    def test_decompress_with_no_data(self):
        """Testing LexersMappingWidget.decompress with no data"""
        my_form = TestForm()
        value = (
            my_form.fields['my_mapping_field']
            .widget
            .decompress(()))
        self.assertEqual(value, [None, None])



"""The app definition for reviewboard.oauth."""

from django.apps import AppConfig


class OAuthAppConfig(AppConfig):
    """App configuration for reviewboard.oauth."""

    name = 'reviewboard.oauth'

    def ready(self):
        """Configure the app once it's ready.

        This will connect signal handlers for the app.

        Version Added:
            5.0
        """
        from reviewboard.oauth.signal_handlers import connect_signal_handlers

        connect_signal_handlers()



"""Add ChangeDescription.user.

Version Added:
    3.0
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('ChangeDescription', 'user', models.ForeignKey, null=True,
             related_model='auth.User'),
]



"""Unit tests for the users list page.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import List, Optional, TYPE_CHECKING, Union

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import AnonymousUser, User
from django.db.models import Count, Q
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.models import Profile
from reviewboard.datagrids.tests.base import BaseViewTestCase
from reviewboard.site.models import LocalSite
from reviewboard.testing.queries.http import get_http_request_start_equeries

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries


class UsersDataGridTests(BaseViewTestCase):
    """Unit tests for the users view."""

    datagrid_url = '/users/'

    @add_fixtures(['test_users'])
    def test_with_access(self):
        """Testing users_list view"""
        self._test_with_access()

    @add_fixtures(['test_site', 'test_users'])
    def test_with_access_with_local_site(self):
        """Testing users_list view on a Local Site"""
        self._test_with_access(with_local_site=True,
                               local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_with_access_with_local_sites_in_db(self):
        """Testing users_list view with Local Sites in the database"""
        self._test_with_access(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_with_access_with_letter(self):
        """Testing users_list view with ?letter="""
        self._test_with_access_with_letter()

    @add_fixtures(['test_site', 'test_users'])
    def test_with_access_with_letter_with_local_site(self):
        """Testing users_list view with ?letter= on a Local Site"""
        self._test_with_access_with_letter(with_local_site=True,
                                           local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_with_access_with_letter_with_local_sites_in_db(self):
        """Testing users_list view with ?letter= with Local Sites in the
        database
        """
        self._test_with_access_with_letter(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_as_anonymous_and_redirect(self):
        """Testing users_list view as anonymous with anonymous
        access disabled
        """
        with self.siteconfig_settings({'auth_require_sitewide_login': True},
                                      reload_settings=False):
            response = self.client.get('/users/')

        self.assertEqual(response.status_code, 302)

    @add_fixtures(['test_users'])
    def test_all_profiles_public(self):
        """Testing UsersDataGrid when all user profiles are public"""
        self._test_all_profiles_public()

    @add_fixtures(['test_site', 'test_users'])
    def test_all_profiles_public_with_local_site(self):
        """Testing UsersDataGrid when all user profiles are public on a
        Local Site
        """
        self._test_all_profiles_public(with_local_site=True,
                                       local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_all_profiles_public_with_local_sites_in_db(self):
        """Testing UsersDataGrid when all user profiles are public with
        Local Sites in the database
        """
        self._test_all_profiles_public(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_all_profiles_public_anonymous(self):
        """Testing UsersDataGrid when all user profiles are public and
        the user is anonymous
        """
        self._test_all_profiles_public_anonymous()

    @add_fixtures(['test_site', 'test_users'])
    def test_all_profiles_public_anonymous_with_local_site(self):
        """Testing UsersDataGrid when all user profiles are public and
        the user is anonymous on a Local Site
        """
        self._test_all_profiles_public_anonymous(with_local_site=True,
                                                 local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_all_profiles_public_anonymous_with_local_sites_in_db(self):
        """Testing UsersDataGrid when all user profiles are public and
        the user is anonymous with Local Sites in the database
        """
        self._test_all_profiles_public_anonymous(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_profile_not_exists(self):
        """Testing UsersDataGrid when a profile does not exist"""
        self._test_profile_not_exists()

    @add_fixtures(['test_site', 'test_users'])
    def test_profile_not_exists_with_local_site(self):
        """Testing UsersDataGrid when a profile does not exist on a Local Site
        """
        self._test_profile_not_exists(with_local_site=True,
                                      local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_profile_not_exists_with_local_sites_in_db(self):
        """Testing UsersDataGrid when a profile does not exist with Local Sites
        in the database
        """
        self._test_profile_not_exists(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_all_profiles_private(self):
        """Testing UsersDataGrid when all user profiles are private"""
        self._test_all_profiles_private()

    @add_fixtures(['test_site', 'test_users'])
    def test_all_profiles_private_with_local_site(self):
        """Testing UsersDataGrid when all user profiles are private on a
        Local Site
        """
        self._test_all_profiles_private(with_local_site=True,
                                        local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_all_profiles_private_with_local_sites_in_db(self):
        """Testing UsersDataGrid when all user profiles are private with
        Local Sites in the database
        """
        self._test_all_profiles_private(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_all_profiles_private_anonymous(self):
        """Testing UsersDataGrid when all users profiles are private and the
        user is anonymous
        """
        self._test_all_profiles_private_anonymous()

    @add_fixtures(['test_site', 'test_users'])
    def test_all_profiles_private_anonymous_with_local_site(self):
        """Testing UsersDataGrid when all users profiles are private and the
        user is anonymous on a Local Site
        """
        self._test_all_profiles_private_anonymous(with_local_site=True,
                                                  local_sites_in_db=True)

    @add_fixtures(['test_site', 'test_users'])
    def test_all_profiles_private_anonymous_with_local_sites_in_db(self):
        """Testing UsersDataGrid when all users profiles are private and the
        user is anonymous with Local Sites in the database
        """
        self._test_all_profiles_private_anonymous(local_sites_in_db=True)

    @add_fixtures(['test_users'])
    def test_all_profiles_private_admin(self):
        """Testing UsersDataGrid when all users profiles are private and the
        user is an admin
        """
        self._test_all_profiles_private_admin()

    @add_fixtures(['test_site', 'test_users'])
    def test_all_profiles_private_admin_with_local_site(self):
        """Testing UsersDataGrid when all users profiles are private and the
        user is an admin on a Local Site
        """
        self._test_all_profiles_private_admin(with_local_site=True,
                                              local_sites_in_db=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_all_profile_private_local_site_admin(self):
        """Testing UsersDataGrid when all profiles are private for a LocalSite
        admin on a non-Local Site
        """
        Profile.objects.create(user_id=3)
        Profile.objects.all().update(is_private=True)
        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        self._prefetch_cached(user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            user_pks=[1, 2, 3, 4],
            local_sites_in_db=False,
            has_pending_count_column=False)

        with assert_queries(equeries):
            response = self.client.get('/users/?columns=fullname')

        self.assertEqual(response.status_code, 200)
        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 4)

        rows_by_username = {
            row['object'].username: row
            for row in datagrid.rows
        }

        self.assertInHTML('<a href="/users/admin/">Admin User</a>',
                          rows_by_username['admin']['cells'][0])
        self.assertInHTML('<a href="/users/doc/">Doc Dwarf</a>',
                          rows_by_username['doc']['cells'][0])
        self.assertInHTML('<a href="/users/dopey/"></a>',
                          rows_by_username['dopey']['cells'][0])
        self.assertInHTML('<a href="/users/grumpy/"></a>',
                          rows_by_username['grumpy']['cells'][0])

    @add_fixtures(['test_users', 'test_site'])
    def test_all_profile_private_local_site_admin_local_site(self):
        """Testing UsersDataGrid when all profiles are private for a LocalSite
        admin on a LocalSite
        """
        dopey = User.objects.get(username='dopey')
        Profile.objects.create(user=dopey)
        Profile.objects.all().update(is_private=True)

        local_site = LocalSite.objects.get(name='local-site-2')

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=False,
            user_pks=[1, 2],
            has_pending_count_column=False)

        with assert_queries(equeries):
            response = self.client.get(
                '/s/local-site-2/users/?columns=fullname')

        self.assertEqual(response.status_code, 200)
        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 2)

        for row in datagrid.rows:
            row_user = row['object']
            self.assertInHTML('<a href="/s/local-site-2/users/%s/">%s</a>'
                              % (row_user.username,
                                 row_user.get_full_name()),
                              row['cells'][0])

    def _test_with_access(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for a user with access to the datagrid.

        This accesses the users datagrid as a logged-in user with access
        to the global site or Local Site (depending on the test), checking
        queries and datagrid row results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        for i in range(5):
            self.create_user(username='test-user-%s' % (i + 1))

        all_users = User.objects.in_bulk()

        if local_site:
            local_site.users.add(*all_users.keys())

        # Make sure every user has a profile for this test.
        for temp_user in all_users.values():
            temp_user.get_profile()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            user_pks=[1, 2, 3, 4, 5, 6, 7, 8, 9])

        with assert_queries(equeries):
            response = self.client.get(
                self.get_datagrid_url(local_site=local_site))

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 9)
        self.assertEqual(datagrid.rows[0]['object'].username, 'admin')
        self.assertEqual(datagrid.rows[1]['object'].username, 'doc')
        self.assertEqual(datagrid.rows[2]['object'].username, 'dopey')
        self.assertEqual(datagrid.rows[3]['object'].username, 'grumpy')
        self.assertEqual(datagrid.rows[4]['object'].username, 'test-user-1')
        self.assertEqual(datagrid.rows[5]['object'].username, 'test-user-2')
        self.assertEqual(datagrid.rows[6]['object'].username, 'test-user-3')
        self.assertEqual(datagrid.rows[7]['object'].username, 'test-user-4')
        self.assertEqual(datagrid.rows[8]['object'].username, 'test-user-5')

    def _test_with_access_with_letter(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for paginating by letter for a user with access.

        This accesses the users datagrid as a logged-in user with access
        to the global site or Local Site (depending on the test), checking
        queries and datagrid row results when filtering by a letter.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        matched_users = [
            self.create_user(username='aa'),
            self.create_user(username='ab'),
            self.create_user(username='ac'),
            User.objects.get(username='admin'),
        ]

        self.create_user(username='da')
        self.create_user(username='db')

        all_users = User.objects.in_bulk()

        # Make sure every user has a profile for this test.
        if local_site:
            local_site.users.add(*all_users.keys())

        for temp_user in all_users.values():
            temp_user.get_profile()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=False,
            user_pks=[
                _user.pk
                for _user in matched_users
            ],
            query=Q(username__istartswith='A'))

        with assert_queries(equeries):
            response = self.client.get(
                '%s?letter=A' % self.get_datagrid_url(local_site=local_site))

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None
        self.assertEqual(len(datagrid.rows), 4)
        self.assertEqual(datagrid.rows[0]['object'].username, 'aa')
        self.assertEqual(datagrid.rows[1]['object'].username, 'ab')
        self.assertEqual(datagrid.rows[2]['object'].username, 'ac')
        self.assertEqual(datagrid.rows[3]['object'].username, 'admin')

    def _test_all_profiles_public(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for access when all profiles are public.

        This accesses the users datagrid as a logged-in user with access
        to the global site or Local Site (depending on the test), checking
        queries and datagrid row results and ensuring public profiles can
        be read.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        datagrid_url = self.get_datagrid_url(local_site=local_site)

        Profile.objects.create(user_id=3)
        self.client.login(username='doc', password='doc')

        if local_site:
            local_site.users.add(*User.objects.all())

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=False,
            user_pks=[1, 2, 3, 4],
            has_pending_count_column=False)

        with assert_queries(equeries):
            response = self.client.get(f'{datagrid_url}?columns=fullname')

        self.assertEqual(response.status_code, 200)
        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 4)

        for row in datagrid.rows:
            row_user = row['object']
            self.assertInHTML('<a href="%s%s/">%s</a>'
                              % (datagrid_url,
                                 row_user.username,
                                 row_user.get_full_name()),
                              row['cells'][0])

    def _test_all_profiles_public_anonymous(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for anonymous access with all public profiles.

        This accesses the users datagrid as an anonymous user with access
        to the global site or a public Local Site (depending on the test),
        checking queries and datagrid row results and ensuring public profiles
        can be read.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        datagrid_url = self.get_datagrid_url(local_site=local_site)

        Profile.objects.create(user_id=3)
        self.client.logout()

        if local_site:
            local_site.public = True
            local_site.save(update_fields=('public',))

            local_site.users.add(*User.objects.all())

        self._prefetch_cached(local_site=local_site)

        equeries = self._build_datagrid_equeries(
            user=AnonymousUser(),
            local_site=local_site,
            local_sites_in_db=False,
            user_pks=[1, 2, 3, 4],
            has_pending_count_column=False)

        with assert_queries(equeries):
            response = self.client.get(f'{datagrid_url}?columns=fullname')

        self.assertEqual(response.status_code, 200)
        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 4)

        for row in datagrid.rows:
            row_user = row['object']
            self.assertInHTML('<a href="%s%s/"></a>' % (datagrid_url,
                                                        row_user.username),
                              row['cells'][0])

    def _test_profile_not_exists(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for access and not all user profiles existing.

        This accesses the users datagrid as a logged-in user with access to the
        global site or a Local Site (depending on the test), checking queries
        and datagrid row results and ensuring missing profiles return the
        expected results.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        datagrid_url = self.get_datagrid_url(local_site=local_site)

        Profile.objects.all().update(is_private=True)
        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        dopey = User.objects.get(username='dopey')

        # Remove doc's admin status from other test Local Sites, so he's
        # never an admin for any other users. We want to test with minimal
        # visibility.
        user.local_site_admins.clear()

        if local_site:
            local_site.users.add(*User.objects.all())

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            user_pks=[1, 2, 3, 4],
            has_pending_count_column=False)
        equeries += [
            {
                'model': Profile,
                'where': Q(user=dopey),
            },
            {
                'model': Profile,
                'type': 'INSERT',
            },
        ]

        with assert_queries(equeries):
            response = self.client.get(f'{datagrid_url}?columns=fullname')

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 4)

        rows_by_username = {
            row['object'].username: row
            for row in datagrid.rows
        }

        self.assertHTMLEqual(
            rows_by_username['admin']['cells'][0],
            f'<td class="has-link" colspan="2">'
            f' <a href="{datagrid_url}admin/"></a>'
            f'</td>')
        self.assertHTMLEqual(
            rows_by_username['doc']['cells'][0],
            f'<td class="has-link" colspan="2">'
            f' <a href="{datagrid_url}doc/">Doc Dwarf</a>'
            f'</td>')
        self.assertHTMLEqual(
            rows_by_username['dopey']['cells'][0],
            f'<td class="has-link" colspan="2">'
            f' <a href="{datagrid_url}dopey/">Dopey Dwarf</a>'
            f'</td>')
        self.assertHTMLEqual(
            rows_by_username['grumpy']['cells'][0],
            f'<td class="has-link" colspan="2">'
            f' <a href="{datagrid_url}grumpy/"></a>'
            f'</td>')

    def _test_all_profiles_private(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for access and all private user profiles.

        This accesses the users datagrid as a logged-in user with access to the
        global site or a Local Site (depending on the test), checking queries
        and datagrid row results and that no private user profile data (aside
        from the user's) can be read.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        datagrid_url = self.get_datagrid_url(local_site=local_site)

        Profile.objects.create(user_id=3)
        Profile.objects.all().update(is_private=True)

        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # Remove doc's admin status from other test Local Sites, so he's
        # never an admin for any other users. We want to test with minimal
        # visibility.
        user.local_site_admins.clear()

        if local_site:
            local_site.users.add(*User.objects.all())

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=False,
            user_pks=[1, 2, 3, 4],
            has_pending_count_column=False)

        with assert_queries(equeries):
            response = self.client.get(f'{datagrid_url}?columns=fullname')

        self.assertEqual(response.status_code, 200)

        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 4)

        rows_by_username = {
            row['object'].username: row
            for row in datagrid.rows
        }

        self.assertHTMLEqual(
            rows_by_username['admin']['cells'][0],
            f'<td class="has-link" colspan="2">'
            f' <a href="{datagrid_url}admin/"></a>'
            f'</td>')
        self.assertHTMLEqual(
            rows_by_username['doc']['cells'][0],
            f'<td class="has-link" colspan="2">'
            f' <a href="{datagrid_url}doc/">Doc Dwarf</a>'
            f'</td>')
        self.assertHTMLEqual(
            rows_by_username['dopey']['cells'][0],
            f'<td class="has-link" colspan="2">'
            f' <a href="{datagrid_url}dopey/"></a>'
            f'</td>')
        self.assertHTMLEqual(
            rows_by_username['grumpy']['cells'][0],
            f'<td class="has-link" colspan="2">'
            f' <a href="{datagrid_url}grumpy/"></a>'
            f'</td>')

    def _test_all_profiles_private_anonymous(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for anonymous access and all private user profiles.

        This accesses the users datagrid as an anonymous user with access to
        the global site or a public Local Site (depending on the test),
        checking queries and datagrid row results and that no private user
        profile data can be read.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        datagrid_url = self.get_datagrid_url(local_site=local_site)

        Profile.objects.create(user_id=3)
        Profile.objects.all().update(is_private=True)
        self.client.logout()

        if local_site:
            local_site.public = True
            local_site.save(update_fields=('public',))
            local_site.users.add(*User.objects.all())

        self._prefetch_cached(local_site=local_site)

        equeries = self._build_datagrid_equeries(
            user=AnonymousUser(),
            local_site=local_site,
            local_sites_in_db=False,
            user_pks=[1, 2, 3, 4],
            has_pending_count_column=False)

        with assert_queries(equeries):
            response = self.client.get(f'{datagrid_url}?columns=fullname')

        self.assertEqual(response.status_code, 200)
        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 4)

        for row in datagrid.rows:
            row_user = row['object']
            self.assertInHTML('<a href="%s%s/"></a>' % (datagrid_url,
                                                        row_user.username),
                              row['cells'][0])

    def _test_all_profiles_private_admin(
        self,
        *,
        with_local_site: bool = False,
        local_sites_in_db: bool = False,
    ) -> None:
        """Common tests for admin access and all private user profiles.

        This accesses the users datagrid as an admin user, checking queries
        and datagrid row results and ensuring all private profiles can be
        read.

        Args:
            with_local_site (bool, optional):
                Whether to test with a Local Site for the query and objects.

            local_sites_in_db (bool, optional):
                Whether to expect Local Sites in the database.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        local_site: Optional[LocalSite]

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        datagrid_url = self.get_datagrid_url(local_site=local_site)

        Profile.objects.create(user_id=3)
        Profile.objects.all().update(is_private=True)
        self.client.login(username='admin', password='admin')

        user = User.objects.get(username='admin')
        profile = user.get_profile()

        if local_site:
            local_site.users.add(*User.objects.all())

        self._prefetch_cached(local_site=local_site,
                              user=user)

        equeries = self._build_datagrid_equeries(
            user=user,
            profile=profile,
            local_site=local_site,
            local_sites_in_db=local_sites_in_db,
            user_pks=[1, 2, 3, 4],
            has_pending_count_column=False)

        with assert_queries(equeries):
            response = self.client.get(f'{datagrid_url}?columns=fullname')

        self.assertEqual(response.status_code, 200)
        datagrid = self._get_context_var(response, 'datagrid')
        assert datagrid is not None

        self.assertEqual(len(datagrid.rows), 4)

        self.assertInHTML(
            f'<a href="{datagrid_url}admin/">Admin User</a>',
            datagrid.rows[0]['cells'][0])
        self.assertInHTML(
            f'<a href="{datagrid_url}doc/">Doc Dwarf</a>',
            datagrid.rows[1]['cells'][0])
        self.assertInHTML(
            f'<a href="{datagrid_url}dopey/">Dopey Dwarf</a>',
            datagrid.rows[2]['cells'][0])
        self.assertInHTML(
            f'<a href="{datagrid_url}grumpy/">Grumpy Dwarf</a>',
            datagrid.rows[3]['cells'][0])

    def _build_datagrid_equeries(
        self,
        *,
        user: Union[AnonymousUser, User],
        user_pks: List[int],
        profile: Optional[Profile] = None,
        local_site: Optional[LocalSite] = None,
        local_sites_in_db: bool = False,
        query: Q = Q(),
        has_pending_count_column: bool = True,
    ) -> ExpectedQueries:
        """Return expected queries for viewing the datagrid.

        This assumes that the user has access to the datagrid, and that any
        cacheable state is already cached.

        Version Added:
            5.0.7

        Args:
            user (django.contrib.auth.models.User):
                The user accessing the datagrid.

            user_pks (list of int):
                The list of user IDs that are expected to be listed, in result
                order.

            profile (reviewboard.accounts.models.Profile):
                The user's profile.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site that's being accessed, if any.

            local_sites_in_db (bool, optional):
                Whether the database contains any Local Sites.

            query (django.db.models.Q, optional):
                An optional query for filtering users.

            has_pending_count_column (bool, optional):
                Whether this datagrid is rendering the Pending Review Request
                Count column.

        Returns:
            list of dict:
            The list of expected queries.
        """
        equeries = get_http_request_start_equeries(
            user=user,
            local_site=local_site)

        if user.is_authenticated:
            assert profile is not None

            equeries += [
                {
                    'model': Profile,
                    'type': 'UPDATE',
                    'where': Q(pk=profile.pk),
                },
            ]

        if local_site:
            equeries += [
                {
                    '__note__': (
                        'Fetch the number of items across all datagrid pages '
                        'on a Local Site'
                    ),
                    'annotations': {'__count': Count('*')},
                    'join_types': {
                        'site_localsite_users': 'INNER JOIN',
                    },
                    'model': User,
                    'num_joins': 1,
                    'tables': {
                        'auth_user',
                        'site_localsite_users',
                    },
                    'where': (Q(local_site=local_site) &
                              query &
                              Q(is_active=True)),
                },
                {
                    '__note__': (
                        'Fetch the IDs of the items for one page on a '
                        'Local Site'
                    ),
                    'distinct': True,
                    'limit': len(user_pks),
                    'join_types': {
                        'site_localsite_users': 'INNER JOIN',
                    },
                    'model': User,
                    'num_joins': 1,
                    'order_by': ('username',),
                    'tables': {
                        'auth_user',
                        'site_localsite_users',
                    },
                    'values_select': ('pk',),
                    'where': (Q(local_site=local_site) &
                              query &
                              Q(is_active=True)),
                },
            ]
        else:
            equeries += [
                {
                    '__note__': (
                        'Fetch the number of items across all datagrid pages'
                    ),
                    'annotations': {'__count': Count('*')},
                    'model': User,
                    'where': (query &
                              Q(is_active=True)),
                },
                {
                    '__note__': (
                        'Fetch the IDs of the items for one page on Local Site'
                    ),
                    'distinct': True,
                    'limit': len(user_pks),
                    'model': User,
                    'order_by': ('username',),
                    'values_select': ('pk',),
                    'where': (query &
                              Q(is_active=True)),
                },
            ]

        if has_pending_count_column:
            equeries += [
                {
                    '__note__': (
                        'Fetch the data for one page based on the IDs'
                    ),
                    'annotations': {
                        'column_pending_review_request_count': Count(
                            'review_requests',
                            filter=(Q(review_requests__public=True) &
                                    Q(review_requests__status='P'))),
                    },
                    'group_by': True,
                    'join_types': {
                        'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    },
                    'model': User,
                    'num_joins': 1,
                    'select_related': {
                        'profile',
                    },
                    'tables': {
                        'auth_user',
                        'reviews_reviewrequest',
                    },
                    'where': Q(pk__in=user_pks),
                },
            ]
        else:
            equeries += [
                {
                    '__note__': (
                        'Fetch the data for one page based on the IDs'
                    ),
                    'model': User,
                    'select_related': {
                        'profile',
                    },
                    'where': Q(pk__in=user_pks),
                },
            ]

        return equeries



import logging

from django.conf import settings
from djblets.siteconfig.models import SiteConfiguration

from reviewboard import initialize
from reviewboard.admin.checks import check_updates_required
from reviewboard.admin.siteconfig import load_site_config
from reviewboard.admin.views import manual_updates_required


logger = logging.getLogger(__name__)


_initialized = False


def init_review_board_middleware(get_response):
    """Handle the initialization of Review Board.

    Args:
        get_response (callable):
            The method to execute the view.
    """
    def middleware(request):
        """Ensure that Review Board initialization code has run.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            django.http.HttpResponse:
            The response object.
        """
        global _initialized

        if not _initialized:
            initialize()
            _initialized = True

        return get_response(request)

    return middleware


def load_settings_middleware(get_response):
    """Middleware that loads the settings on each request.

    Args:
        get_response (callable):
            The method to execute the view.
    """
    def middleware(request):
        """Ensure that the latest siteconfig is loaded.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            django.http.HttpResponse:
            The response object.
        """
        try:
            siteconfig = SiteConfiguration.objects.get_current()
        except Exception as e:
            logger.critical('Unable to load SiteConfiguration: %s',
                            e, exc_info=True, extra={'request': request})
            return

        # This will be unset if the SiteConfiguration expired, since we'll
        # have a new one in the cache.
        if not hasattr(siteconfig, '_rb_settings_loaded'):
            # Load all site settings.
            load_site_config()
            siteconfig._rb_settings_loaded = True

        if siteconfig.settings.get('site_domain_method', 'http') == 'https':
            request.META['wsgi.url_scheme'] = 'https'

        return get_response(request)

    return middleware


def check_updates_required_middleware(get_response):
    """Middleware that checks if manual updates need to be done.

    If updates are required, all attempts to access a URL will be redirected to
    the updates page (or an appropriate error response for API calls).

    Args:
        get_response (callable):
            The method to execute the view.
    """
    ALLOWED_PATHS = (
        settings.STATIC_URL,
        settings.SITE_ROOT + 'jsi18n/',
    )

    def middleware(request):
        """Check whether updates are required.

        This returns the appropriate response if any updates are required,
        otherwise it allows the normal code to run.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            django.http.HttpResponse:
            The response object.
        """
        path_info = request.META['PATH_INFO']

        updates_required = check_updates_required()

        if updates_required and not path_info.startswith(ALLOWED_PATHS):
            return manual_updates_required(request, updates_required)

        return get_response(request)

    return middleware


class ExtraExceptionInfoMiddleware(object):
    """Add extra debugging information to exception e-mails.

    If an exception occurs, the META field will be updated to contain
    the username and e-mail address of the user who triggered the error
    (if any), and the Local Site name (if any).
    """

    def __init__(self, get_response):
        """Initialize the middleware.

        Args:
            get_response (callable):
                The method to call to get the response object.
        """
        self.get_response = get_response

    def process_exception(self, request, exception):
        """Process an exception.

        Exceptions from views are handled by sending the admin users an e-mail
        with the traceback. This adds additional information to the META
        dictionary before that happens.
        """
        if request.user.is_authenticated:
            request.META['USERNAME'] = request.user.username
            request.META['USER_EMAIL'] = request.user.email

        if hasattr(request, '_local_site_name'):
            request.META['LOCAL_SITE'] = request._local_site_name

    def __call__(self, request):
        """Run the middleware.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            django.http.HttpResponse:
            The response object.
        """
        return self.get_response(request)



"""Unit tests for reviewboard.reviews.ui.base.ReviewUI."""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from django.utils.safestring import SafeText
from djblets.testing.decorators import add_fixtures
from kgb import SpyAgency

from reviewboard.reviews.models import FileAttachmentComment
from reviewboard.reviews.ui.base import ReviewUI
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from reviewboard.reviews.models.base_comment import BaseComment


class DummyReviewableObject(object):
    pass


class MyReviewUI(ReviewUI):
    """A basic file attachment Review UI used for testing."""

    supported_mimetypes = ['application/rbtest']
    supports_diffing = True
    supports_file_attachments = True
    js_model_class: str = 'RB.Test.ReviewUI'
    js_view_class: str = 'RB.Test.ReviewUIView'
    js_bundle_names = [
        'admin',
    ]
    css_bundle_names = [
        'js-tests',
    ]

    def get_caption(self, draft=None) -> str:
        return 'Test Caption'

    def get_comments(self) -> list[BaseComment]:
        return list(FileAttachmentComment.objects.order_by('pk'))

    def get_page_cover_image_url(self) -> str:
        return '/cover-image.png'

    def get_extra_context(self, request) -> dict[str, Any]:
        return {
            'custom_key': '123',
        }


class ReviewUITests(SpyAgency, TestCase):
    """Unit tests for reviewboard.reviews.ui.base.ReviewUI."""

    fixtures = ['test_users']

    def setUp(self) -> None:
        super().setUp()

        self.review_request = self.create_review_request()
        self.attachment = self.create_file_attachment(self.review_request)
        self.review = self.create_review(self.review_request, public=False)
        self.create_file_attachment_comment(self.review,
                                            self.attachment,
                                            text='Comment 1')
        self.create_file_attachment_comment(self.review,
                                            self.attachment,
                                            text='Comment 2',
                                            issue_opened=True)

    def test_build_render_context(self) -> None:
        """Testing ReviewUI.build_render_context"""
        reviewable_obj1 = DummyReviewableObject()
        reviewable_obj2 = DummyReviewableObject()

        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=reviewable_obj2)
        review_ui.set_diff_against(reviewable_obj1)

        request = self.create_http_request(user=self.review.user)
        context = review_ui.build_render_context(request=request)

        self.assertEqual(context['base_template'], 'reviews/ui/base.html')
        self.assertEqual(context['caption'], 'Test Caption')
        self.assertEqual(context['social_page_image_url'], '/cover-image.png')
        self.assertEqual(context['social_page_title'],
                         'Reviewable for Review Request #1: Test Caption')
        self.assertEqual(context['review_request'], self.review_request)
        self.assertEqual(context['review_request_details'],
                         self.review_request)
        self.assertEqual(context['review'], self.review)
        self.assertFalse(context['review_ui_inline'])
        self.assertFalse(context['skip_static_media'])
        self.assertFalse(context['draft'])
        self.assertIs(context['review_ui'], review_ui)
        self.assertIs(context['obj'], reviewable_obj2)
        self.assertIs(context['diff_against_obj'], reviewable_obj1)
        self.assertIn('close_description', context)
        self.assertIn('close_description_rich_text', context)
        self.assertIn('last_activity_time', context)
        self.assertIn('review_ui_uuid', context)
        self.assertEqual(context['custom_key'], '123')

        self.assertEqual(len(context['comments']), 2)
        self.assertEqual(context['comments'][0].text, 'Comment 1')
        self.assertEqual(context['comments'][1].text, 'Comment 2')

    def test_build_render_context_with_inline_true(self) -> None:
        """Testing ReviewUI.build_render_context with inline=1"""
        reviewable_obj1 = DummyReviewableObject()
        reviewable_obj2 = DummyReviewableObject()

        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=reviewable_obj2)
        review_ui.set_diff_against(reviewable_obj1)

        request = self.create_http_request()
        context = review_ui.build_render_context(request=request,
                                                 inline=True)

        self.assertEqual(context['base_template'],
                         'reviews/ui/base_inline.html')
        self.assertEqual(context['caption'], 'Test Caption')
        self.assertEqual(context['social_page_image_url'], '/cover-image.png')
        self.assertEqual(context['social_page_title'],
                         'Reviewable for Review Request #1: Test Caption')
        self.assertEqual(context['review_request'], self.review_request)
        self.assertEqual(context['review_request_details'],
                         self.review_request)
        self.assertTrue(context['review_ui_inline'])
        self.assertFalse(context['skip_static_media'])
        self.assertFalse(context['draft'])
        self.assertIs(context['review_ui'], review_ui)
        self.assertIs(context['obj'], reviewable_obj2)
        self.assertIs(context['diff_against_obj'], reviewable_obj1)
        self.assertIn('close_description', context)
        self.assertIn('close_description_rich_text', context)
        self.assertIn('last_activity_time', context)
        self.assertIn('social_page_image_url', context)
        self.assertIn('review_ui_uuid', context)
        self.assertNotIn('review', context)
        self.assertEqual(context['custom_key'], '123')

        self.assertEqual(len(context['comments']), 2)
        self.assertEqual(context['comments'][0].text, 'Comment 1')
        self.assertEqual(context['comments'][1].text, 'Comment 2')

    def test_build_render_context_with_skip_static_media_true(self) -> None:
        """Testing ReviewUI.build_render_context with skip_static_media=1"""
        reviewable_obj1 = DummyReviewableObject()
        reviewable_obj2 = DummyReviewableObject()

        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=reviewable_obj2)
        review_ui.set_diff_against(reviewable_obj1)

        request = self.create_http_request(data={
            'skip-static-media': '1',
        })
        context = review_ui.build_render_context(request=request,
                                                 inline=True)

        self.assertEqual(context['base_template'],
                         'reviews/ui/base_inline.html')
        self.assertEqual(context['caption'], 'Test Caption')
        self.assertEqual(context['social_page_image_url'], '/cover-image.png')
        self.assertEqual(context['social_page_title'],
                         'Reviewable for Review Request #1: Test Caption')
        self.assertEqual(context['review_request'], self.review_request)
        self.assertEqual(context['review_request_details'],
                         self.review_request)
        self.assertTrue(context['review_ui_inline'])
        self.assertTrue(context['skip_static_media'])
        self.assertFalse(context['draft'])
        self.assertIs(context['review_ui'], review_ui)
        self.assertIs(context['obj'], reviewable_obj2)
        self.assertIs(context['diff_against_obj'], reviewable_obj1)
        self.assertIn('close_description', context)
        self.assertIn('close_description_rich_text', context)
        self.assertIn('last_activity_time', context)
        self.assertIn('social_page_image_url', context)
        self.assertIn('review_ui_uuid', context)
        self.assertNotIn('review', context)
        self.assertEqual(context['custom_key'], '123')

        self.assertEqual(len(context['comments']), 2)
        self.assertEqual(context['comments'][0].text, 'Comment 1')
        self.assertEqual(context['comments'][1].text, 'Comment 2')

    def test_get_comments_json(self) -> None:
        """Testing ReviewUI.get_comments_json"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        review_ui.request = self.create_http_request(user=self.review.user)

        comments_json = review_ui.get_comments_json()

        self.assertJSONEqual(
            comments_json,
            {
                '0': [
                    {
                        'comment_id': 1,
                        'html': 'Comment 1',
                        'issue_opened': False,
                        'issue_status': '',
                        'localdraft': True,
                        'reply_to_id': None,
                        'review_id': 1,
                        'review_request_id': 1,
                        'rich_text': False,
                        'text': 'Comment 1',
                        'url': '/r/1/#fcomment1',
                        'user': {
                            'name': 'Dopey Dwarf',
                            'username': 'dopey',
                        },
                    },
                ],
                '1': [
                    {
                        'comment_id': 2,
                        'html': 'Comment 2',
                        'issue_opened': True,
                        'issue_status': 'open',
                        'localdraft': True,
                        'reply_to_id': None,
                        'review_id': 1,
                        'review_request_id': 1,
                        'rich_text': False,
                        'text': 'Comment 2',
                        'url': '/r/1/#fcomment2',
                        'user': {
                            'name': 'Dopey Dwarf',
                            'username': 'dopey',
                        },
                    },
                ],
            })

    def test_render_to_response_with_inline_false(self) -> None:
        """Testing ReviewUI.render_to_response with inline=0"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        request = self.create_http_request(data={
            'inline': '0',
        })

        response = review_ui.render_to_response(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b'model: new RB.Test.ReviewUI(',
                      response.content)
        self.assertIn(b'view = new RB.Test.ReviewUIView(',
                      response.content)
        self.assertIn(b'"renderedInline": false', response.content)

    def test_render_to_response_with_inline_true(self) -> None:
        """Testing ReviewUI.render_to_response with inline=1"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        request = self.create_http_request(data={
            'inline': '1',
        })

        response = review_ui.render_to_response(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b'model: new RB.Test.ReviewUI(',
                      response.content)
        self.assertIn(b'view = new RB.Test.ReviewUIView(',
                      response.content)
        self.assertIn(b'"renderedInline": true', response.content)

    def test_render_to_string_with_inline_false(self) -> None:
        """Testing ReviewUI.render_to_string with inline=False"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        request = self.create_http_request()

        content = review_ui.render_to_string(request, inline=False)
        self.assertIsInstance(content, SafeText)
        self.assertIn('model: new RB.Test.ReviewUI(', content)
        self.assertIn('view = new RB.Test.ReviewUIView(', content)
        self.assertIn('"renderedInline": false', content)

    def test_render_to_string_with_inline_true(self) -> None:
        """Testing ReviewUI.render_to_string with inline=True"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        request = self.create_http_request()

        content = review_ui.render_to_string(request, inline=True)
        self.assertIsInstance(content, SafeText)
        self.assertIn('model: new RB.Test.ReviewUI(', content)
        self.assertIn('view = new RB.Test.ReviewUIView(', content)
        self.assertIn('"renderedInline": true', content)

    def test_render_to_response_with_skip_static_media_false(self) -> None:
        """Testing ReviewUI.render_to_response with skip_static_media=0"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        request = self.create_http_request(data={
            'skip-static-media': '0',
        })

        response = review_ui.render_to_response(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(
            b'link href="/static/rb/css/js-tests.min.css" '
            b'rel="stylesheet" type="text/css"',
            response.content)
        self.assertIn(
            b'script type="text/javascript" src="/static/rb/js/admin.min.js"',
            response.content)

    def test_render_to_response_with_skip_static_media_true(self) -> None:
        """Testing ReviewUI.render_to_response with skip_static_media=1"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        request = self.create_http_request(data={
            'skip-static-media': '1',
        })

        response = review_ui.render_to_response(request)
        self.assertEqual(response.status_code, 200)
        self.assertNotIn(
            b'link href="/static/rb/css/js-tests.min.css" '
            b'rel="stylesheet" type="text/css"',
            response.content)
        self.assertNotIn(
            b'script type="text/javascript" src="/static/rb/js/admin.min.js"',
            response.content)

    def test_render_to_string_with_skip_static_media_false(self) -> None:
        """Testing ReviewUI.render_to_response with skip_static_media=0"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        request = self.create_http_request(data={
            'skip-static-media': '0',
        })

        content = review_ui.render_to_string(request)
        self.assertIn(
            'link href="/static/rb/css/js-tests.min.css" '
            'rel="stylesheet" type="text/css"',
            content)
        self.assertIn(
            'script type="text/javascript" src="/static/rb/js/admin.min.js"',
            content)

    def test_render_to_string_with_skip_static_media_true(self) -> None:
        """Testing ReviewUI.render_to_response with skip_static_media=1"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        request = self.create_http_request(data={
            'skip-static-media': '1',
        })

        content = review_ui.render_to_string(request)
        self.assertNotIn(
            'link href="/static/rb/css/js-tests.min.css" '
            'rel="stylesheet" type="text/css"',
            content)
        self.assertNotIn(
            'script type="text/javascript" src="/static/rb/js/admin.min.js"',
            content)

    def test_serialize_comments(self) -> None:
        """Testing ReviewUI.serialize_comments"""
        # Note that this test is factoring in both the user-owned draft
        # review created in setUp() and the ones created in this test.
        #
        # Comments from review1 and review2 will be included in the results.
        review1 = self.review

        review2 = self.create_review(self.review_request)
        self.create_file_attachment_comment(review2, self.attachment,
                                            text='Comment 3')

        # This will not be included.
        review3 = self.create_review(self.review_request,
                                     user=self.create_user('user2'),
                                     public=False)
        self.create_file_attachment_comment(review3, self.attachment,
                                            text='Comment 4')

        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        review_ui.request = self.create_http_request(user=review1.user)

        comments = review_ui.serialize_comments(list(review_ui.get_comments()))

        self.assertEqual(len(comments), 3)
        self.assertEqual(comments['0'][0]['text'], 'Comment 1')
        self.assertEqual(comments['1'][0]['text'], 'Comment 2')
        self.assertEqual(comments['2'][0]['text'], 'Comment 3')

    def test_serialize_comment(self) -> None:
        """Testing ReviewUI.serialize_comment"""
        review_ui = MyReviewUI(review_request=self.review_request,
                               obj=DummyReviewableObject())
        review_ui.request = self.create_http_request(user=self.review.user)

        self.assertEqual(
            review_ui.serialize_comment(review_ui.get_comments()[0]),
            {
                'comment_id': 1,
                'html': 'Comment 1',
                'issue_opened': False,
                'issue_status': '',
                'localdraft': True,
                'reply_to_id': None,
                'review_id': 1,
                'review_request_id': 1,
                'rich_text': False,
                'text': 'Comment 1',
                'url': '/r/1/#fcomment1',
                'user': {'name': 'Dopey Dwarf', 'username': 'dopey'}
            })

    @add_fixtures(['test_scmtools'])
    def test_get_comment_link_url_diff(self) -> None:
        """Testing ReviewUI.get_comment_link_url with a file attachment
        attached to a diff
        """
        review_request = self.create_review_request(
            publish=True,
            create_repository=True)
        diffset = self.create_diffset(review_request=review_request)
        filediff = self.create_filediff(diffset=diffset, binary=True)

        attachment = self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=filediff)

        review = self.create_review(review_request=review_request)
        comment = self.create_file_attachment_comment(
            review=review,
            file_attachment=attachment)

        review_ui = MyReviewUI(
            review_request=review_request,
            obj=attachment)

        self.assertEqual(review_ui.get_comment_link_url(comment),
                         f'/r/2/diff/1/#file{filediff.pk}')

    @add_fixtures(['test_scmtools'])
    def test_get_comment_link_url_interdiff(self) -> None:
        """Testing ReviewUI.get_comment_link_url with a file attachment
        attached to an interdiff
        """
        review_request = self.create_review_request(
            publish=True,
            create_repository=True)

        diffset1 = self.create_diffset(review_request=review_request)
        filediff1 = self.create_filediff(diffset=diffset1, binary=True)
        attachment1 = self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=filediff1)

        diffset2 = self.create_diffset(
            review_request=review_request,
            revision=2)
        filediff2 = self.create_filediff(diffset=diffset2, binary=True)
        attachment2 = self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=filediff2)

        review = self.create_review(review_request=review_request)
        comment = self.create_file_attachment_comment(
            review=review,
            file_attachment=attachment2,
            diff_against_file_attachment=attachment1)

        review_ui = MyReviewUI(
            review_request=review_request,
            obj=attachment2)

        self.assertEqual(review_ui.get_comment_link_url(comment),
                         f'/r/2/diff/1-2/#file{filediff2.pk}')

    @add_fixtures(['test_scmtools'])
    def test_get_comment_link_diff_with_commits(self) -> None:
        """Testing ReviewUI.get_comment_link_url with a file attachment
        attached to a diff that has commit history
        """
        review_request = self.create_review_request(
            publish=True, create_repository=True)

        diffset = self.create_diffset(review_request=review_request)

        cumulative_filediff = self.create_filediff(
            diffset=diffset,
            binary=True)

        commit = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False)
        commit_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit)

        attachment = self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit_filediff)

        review = self.create_review(review_request=review_request)
        comment = self.create_file_attachment_comment(
            review=review,
            file_attachment=attachment)

        review_ui = MyReviewUI(
            review_request=review_request,
            obj=attachment)

        self.assertEqual(review_ui.get_comment_link_url(comment),
                         f'/r/2/diff/1/#file{cumulative_filediff.pk}')

    @add_fixtures(['test_scmtools'])
    def test_get_comment_link_diff_with_commits_base_commit_id(self) -> None:
        """Testing ReviewUI.get_comment_link_url with a file attachment
        attached to a diff that has commit history with a specified base commit
        """
        review_request = self.create_review_request(
            publish=True, create_repository=True)

        diffset = self.create_diffset(review_request=review_request)

        # Cumulative filediff
        self.create_filediff(
            diffset=diffset,
            binary=True)

        commit1 = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False,
            commit_id='r1',
            parent_id='r0')
        commit1_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit1)
        self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit1_filediff)

        commit2 = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False,
            commit_id='r2',
            parent_id='r1')
        commit2_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit2)
        attachment2 = self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit2_filediff)

        commit3 = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False,
            commit_id='r3',
            parent_id='r2')
        commit3_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit3)
        attachment3 = self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit3_filediff)

        review = self.create_review(review_request=review_request)
        comment = self.create_file_attachment_comment(
            review=review,
            file_attachment=attachment3,
            diff_against_file_attachment=attachment2)

        review_ui = MyReviewUI(
            review_request=review_request,
            obj=attachment3)

        self.assertEqual(review_ui.get_comment_link_url(comment),
                         f'/r/2/diff/1/?base-commit-id={commit2.pk}'
                         f'#file{commit3_filediff.pk}')

    @add_fixtures(['test_scmtools'])
    def test_get_comment_link_diff_with_commits_tip_commit_id(self) -> None:
        """Testing ReviewUI.get_comment_link_url with a file attachment
        attached to a diff that has commit history with a specified tip commit
        """
        review_request = self.create_review_request(
            publish=True, create_repository=True)

        diffset = self.create_diffset(review_request=review_request)

        # Cumulative filediff
        self.create_filediff(
            diffset=diffset,
            binary=True)

        upstream_attachment = self.create_file_attachment(
            review_request=review_request,
            repository=review_request.repository,
            with_history=False,
            repo_path='/test-file',
            repo_revision='r0')

        commit1 = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False,
            commit_id='r1',
            parent_id='r0')
        commit1_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit1)
        self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit1_filediff)

        commit2 = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False,
            commit_id='r2',
            parent_id='r1')
        commit2_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit2)
        attachment2 = self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit2_filediff)

        commit3 = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False,
            commit_id='r3',
            parent_id='r2')
        commit3_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit3)
        self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit3_filediff)

        review = self.create_review(review_request=review_request)
        comment = self.create_file_attachment_comment(
            review=review,
            file_attachment=attachment2,
            diff_against_file_attachment=upstream_attachment)

        review_ui = MyReviewUI(
            review_request=review_request,
            obj=attachment2)

        self.assertEqual(review_ui.get_comment_link_url(comment),
                         f'/r/2/diff/1/?tip-commit-id={commit2.pk}'
                         f'#file{commit2_filediff.pk}')

    @add_fixtures(['test_scmtools'])
    def test_get_comment_link_diff_with_commits_base_and_tip_commit_id(
        self,
    ) -> None:
        """Testing ReviewUI.get_comment_link_url with a file attachment
        attached to a diff that has commit history with a specified base and
        tip commit
        """
        review_request = self.create_review_request(
            publish=True, create_repository=True)

        diffset = self.create_diffset(review_request=review_request)

        # Cumulative filediff
        self.create_filediff(
            diffset=diffset,
            binary=True)

        # Upstream attachment
        self.create_file_attachment(
            review_request=review_request,
            repository=review_request.repository,
            with_history=False,
            repo_path='/test-file',
            repo_revision='r0')

        commit1 = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False,
            commit_id='r1',
            parent_id='r0')
        commit1_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit1)
        attachment1 = self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit1_filediff)

        commit2 = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False,
            commit_id='r2',
            parent_id='r1')
        commit2_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit2)
        attachment2 = self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit2_filediff)

        commit3 = self.create_diffcommit(
            repository=review_request.repository,
            diffset=diffset,
            with_diff=False,
            commit_id='r3',
            parent_id='r2')
        commit3_filediff = self.create_filediff(
            diffset=diffset,
            binary=True,
            commit=commit3)
        self.create_file_attachment(
            review_request=review_request,
            with_history=False,
            added_in_filediff=commit3_filediff)

        review = self.create_review(review_request=review_request)
        comment = self.create_file_attachment_comment(
            review=review,
            file_attachment=attachment2,
            diff_against_file_attachment=attachment1)

        review_ui = MyReviewUI(
            review_request=review_request,
            obj=attachment2)

        self.assertEqual(review_ui.get_comment_link_url(comment),
                         f'/r/2/diff/1/?base-commit-id={commit1.pk}'
                         f'&tip-commit-id={commit2.pk}'
                         f'#file{commit2_filediff.pk}')



"""Unit tests for reviewboard.extensions.hooks.SCMToolHook."""

from reviewboard.extensions.hooks import SCMToolHook
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase
from reviewboard.scmtools import scmtools_registry
from reviewboard.scmtools.core import SCMTool


class TestSCMTool(SCMTool):
    name = 'Dummy'
    scmtool_id = 'dummy-scmtool'


class TestSCMToolNoID(SCMTool):
    name = 'Dummy'
    scmtool_id = None


class SCMToolHookTests(BaseExtensionHookTestCase):
    """Testing SCMToolHook."""

    def test_register(self):
        """Testing SCMToolHook initializing"""
        hook = SCMToolHook(self.extension, TestSCMTool)

        self.assertIs(scmtools_registry.get_by_id('dummy-scmtool'),
                      TestSCMTool)

        hook.disable_hook()

    def test_register_without_scmtool_id(self):
        """Testing SCMToolHook initializing without scmtool_id"""
        message = 'TestSCMToolNoID.scmtool_id must be set.'

        with self.assertRaisesMessage(ValueError, message):
            SCMToolHook(self.extension, TestSCMToolNoID)

    def test_unregister(self):
        """Testing SCMToolHook uninitializing"""
        hook = SCMToolHook(self.extension, TestSCMTool)
        hook.disable_hook()

        self.assertIsNone(scmtools_registry.get_by_id('dummy-scmtool'))



from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _


def validate_repositories(form, field='repositories'):
    """Validate that the repositories all have valid, matching LocalSites.

    This will compare the LocalSite associated with the form to that of each
    added Repository. Each Repository must have the same LocalSite that the
    form is using.

    Args:
        form (django.forms.Form):
            The form to validate.

        field (unicode):
            The name of the form field to validate.

    Returns:
        list of reviewboard.scmtools.models.Repository:
        The list of repositories.

    Raises:
        django.core.exceptions.ValidationError:
            The selected groups contained ones which were not properly limited
            to the local site.
    """
    repositories = form.cleaned_data.get(field, [])
    local_site = form.cleaned_data['local_site']

    for repository in repositories:
        if repository.local_site != local_site:
            raise ValidationError(
                [_("The repository '%s' doesn't exist on the local site.")
                 % repository.name])

    return repositories


def validate_review_groups(form, field='review_groups'):
    """Validate that the review groups all have valid, matching LocalSites.

    This will compare the LocalSite associated with the form to that of
    each added Group. Each Group must have the same LocalSite that the form
    is using.

    Args:
        form (django.forms.Form):
            The form to validate.

        field (unicode):
            The name of the form field to validate.

    Returns:
        list of reviewboard.reviews.models.group.Group:
        The list of groups.

    Raises:
        django.core.exceptions.ValidationError:
            The selected groups contained ones which were not properly limited
            to the local site.
    """
    groups = form.cleaned_data.get(field, [])
    local_site = form.cleaned_data['local_site']

    for group in groups:
        if group.local_site != local_site:
            raise ValidationError(
                [_("The review group %s does not exist.") % group.name])

    return groups


def validate_users(form, field='users'):
    """Validate that the users all have valid, matching LocalSites.

    This will compare the LocalSite associated with the form to that of
    each added User. If the form has a LocalSite set, then all Users are
    required to be a part of that LocalSite. Otherwise, any User is allowed.

    Args:
        form (django.forms.Form):
            The form to validate.

        field (unicode):
            The name of the form field to validate.

    Returns:
        list of django.contrib.auth.models.User:
        The list of users.

    Raises:
        django.core.exceptions.ValidationError:
            The selected users contained ones which were not properly limited
            to the local site.
    """
    local_site = form.cleaned_data['local_site']
    users = form.cleaned_data.get(field, [])

    if local_site:
        for user in users:
            if not user.local_site.filter(pk=local_site.pk).exists():
                raise ValidationError(
                    [_("The user %s is not a member of this site.")
                     % user.username])

    return users
